unit uDemandaAnioBaseEIndices;

{$MODE Delphi}

interface

uses
  SysUtils,
  uauxiliares,
  xmatdefs,
  math,
  uglobs, uDemandas, unodos, usimplex,
  ufechas,
  ucosa, udatoshorariosdetallados,
  uconstantesSimSEE, uFuentesAleatorias, uFichasLPD;

resourcestring
  rsDemandaGenererada =
    'Demanda generada a partir de un año base y vector de energías anuales';

type
  TDemandaAnioBaseEIndices = class(TDemanda)
  public
    ArchiDatos: TArchiRef;
    fIniAnioBase: TFecha;

    demAnioBase: TDAofNReal; // La demanda hora a hora para un año
    // base de 366 días
    invSDem365, invSDem366: NReal; // 1 / la sumatoria de las demandas horarias
    // en [GWh] del año base considerando 365 y
    // 366 días



    anioIni, anioFin: integer;
    // Primer y último año para el que se especificaron valores
    GWh_anios: TDAofNReal; // la cantidad de GWh demandada para cada año
    // siguiente al de inicio de la simulación

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades: TFichasLPD; nodo: TNodo; falla_profundidad, falla_costo: TDAOfNReal;
      ArchiDatos: string; fuente: TFuenteAleatoria; nombreBorne: string;
      SumarEnergiaHr: boolean; xAnioIni, xAnioFin: integer; xGWh_anios: TDAofNReal;
      icf_Fuente: TFuenteAleatoria; icf_NombreBorne: string);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;

    procedure PrepararMemoria(globs: TGlobs); override;

    function InfoAd_: string; override;
    class function DescClase: string; override;

    function PotenciaFirme: NReal; override;

    //      procedure prepararSim; override;
    procedure prepararPaso_as; override;
    procedure Free; override;
  private
    CacheActivo: boolean;
    procedure InitFromFile;
  end;

procedure AlInicio;
procedure AlFinal;

implementation

//------------------------------------
// Métodos de TDemandaAnioBaseEIndices
//====================================


constructor TDemandaAnioBaseEIndices.Create(capa: integer; nombre: string;
  nacimiento, muerte: TFecha; lpdUnidades: TFichasLPD; nodo: TNodo;
  falla_profundidad, falla_costo: TDAOfNReal; ArchiDatos: string;
  fuente: TFuenteAleatoria; nombreBorne: string; SumarEnergiaHr: boolean;
  xAnioIni, xAnioFin: integer; xGWh_anios: TDAofNReal; icf_Fuente: TFuenteAleatoria;
  icf_NombreBorne: string);
begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades, nodo,
    falla_profundidad, falla_costo,
    fuente, nombreBorne,
    icf_Fuente, icf_NombreBorne, SumarEnergiaHr);

  if FileExists(ArchiDatos) { *Converted from FileExists*  } then
    self.ArchiDatos := TArchiRef.Create(ArchiDatos)
  else
    raise ExceptionFileNotFound.Create(ArchiDatos, 'Creando el actor ' + ClaseNombre);
  AnioIni := xAnioIni;
  AnioFin := xAnioFin;
  GWh_anios := xGWh_anios;
  CacheActivo := False;
  InitFromFile;
end;

constructor TDemandaAnioBaseEIndices.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  f.rdArchRef('ArchiDatos', archiDatos);
  if f.Version < 12 then
  begin
    AnioIni := -1;
    AnioFin := -1;
    f.rd('indices', GWh_anios);
  end
  else
  begin
    f.rd('AnioIni', AnioIni);
    f.rd('AnioFin', AnioFin);
    f.rd('GWh_anios', GWh_anios);
  end;

  f.EjecutarLectura;
  CacheActivo := False;
  if not ArchiDatos.testear then
    raise ExceptionFileNotFound.Create(ArchiDatos.archi, 'Creando el actor ' +
      ClaseNombre);
  InitFromFile;
end;

procedure TDemandaAnioBaseEIndices.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wrArchRef('ArchiDatos', ArchiDatos);
  f.wr('AnioIni', AnioIni);
  f.wr('AnioFin', AnioFin);
  f.wr('GWh_anios', GWh_anios);
end;

procedure TDemandaAnioBaseEIndices.prepararMemoria(globs: TGlobs);

begin
  inherited prepararMemoria(globs);

  InitFromFile;

  if AnioIni < 0 then
  begin
    // esta arreglando el entuerto de la versión anterior a la 12 que no
    // tenía AnioIni y AnioFin

    AnioIni := globs.fechaIniOpt.anio;
    Aniofin := globs.fechaFinOpt.anio;
    if length(GWh_anios) < (AnioFin - AnioIni + 1) then
      AnioFin := AnioIni + high(GWh_anios);
    globs.Alerta('Ojo, revise el actor Demanda: ' + nombre +
      '. Fue transformado desde una versión anterior que no especificaba el año de inicio y de fin.');
  end;

  if (anioIni > globs.fechaIniSim.anio) then
    globs.Alerta('ojo, La demanda (' + nombre +
      ') no cubre el inicio de la Simulación, se usará el primer año para cubrir el hueco.');

  if (anioIni > globs.fechaIniOpt.anio) then
    globs.Alerta('ojo, La demanda (' + nombre +
      ') no cubre el inicio de la Optimización, se usará el primer año para cubrir el hueco.');

  if (globs.fechaFinSim.anio > anioFin) then
    globs.Alerta('ojo, La demanda (' + nombre +
      ') no cubre el final de la Simulación, se usará el último año para cubrir el hueco.');

  if (globs.fechaFinOpt.anio > anioFin) then
    globs.Alerta('ojo, La demanda (' + nombre +
      ') no cubre el final de la Optimización, se usará el último año para cubrir el hueco.');

end;

function TDemandaAnioBaseEIndices.infoAd_: string;
begin
  Result := ArchiDatos.archi;
end;

class function TDemandaAnioBaseEIndices.DescClase: string;
begin
  Result := rsDemandaGenererada;
end;

function TDemandaAnioBaseEIndices.potenciaFirme: NReal;
begin
  Result := paUnidades.nUnidades[0] * vmax(self.PPa);
end;

procedure TDemandaAnioBaseEIndices.prepararPaso_as;
var
  factCrecimiento: NReal;
  kAnio: integer; // indice del año del paso dentro del vector de años especificados
  jHoraDelPaso: integer; // hora desde el inicio delpaso 0 ... durpaso-1
  jHoraDelAnio_ultimaHoraDelAnio: integer;
  // auxiliar, se calcula al actualizar el factor
  jHoraDelAnio_desde, jHoraDelAnio_hasta: integer; // limites del paso

  dtInicioDelAnio: TDateTime;
  rh: NReal;
  ih: integer;
  alfa: NReal;
  P1, P2: NReal;

  procedure ActualizarFactorYUltimaHoraDelAnio(anio: integer);
  begin
    kAnio := anio - AnioIni;

    // si actua esta extensión ya se lo avisamos al preparar memoria.
    if kAnio < 0 then
      kAnio := 0
    else if kAnio > high(GWh_anios) then
      kAnio := high(GWh_anios);

    if not IsLeapYear(anio) then
    begin
      factCrecimiento := GWh_anios[kAnio] * invSDem365;
      jHoraDelAnio_UltimaHoraDelAnio := 365 * 24 - 1;
    end
    else
    begin
      factCrecimiento := GWh_anios[kAnio] * invSDem366;
      jHoraDelAnio_UltimaHoraDelAnio := 366 * 24 - 1;
    end;
  end;

begin
  ActualizarFactorYUltimahoraDelAnio(globs.AnioInicioDelPaso);


  if globs.SalaMinutal then
  begin
    rh:= globs.FechaInicioDelpaso.diasDesdeElInicioDelAnio * 24.0;
    ih:= trunc( rh );
    alfa:= frac( rh );
    P1:=  demAnioBase[ ih ];
    inc( ih );
    if ih <= high( demAnioBase ) then
      P2:= demAnioBase[ ih ]
    else
      P2:= demAnioBase[ 0 ];
    PHoraria_[0] := (P1 *(1-alfa)+P2*alfa) * factCrecimiento;
  end
  else
  begin
    // inicializamos puntero de lectura sobre el anio base.
    jHoraDelAnio_desde := globs.FechaInicioDelpaso.horasDesdeElInicioDelAnio;
    jHoraDelAnio_hasta := ceil( jHoraDelAnio_desde + globs.HorasDelPaso - 1 );

  //Cargamos los datos que vamos a usar en el paso
  if jHoraDelAnio_hasta <= jHoraDelAnio_UltimaHoraDelAnio then
  begin // el paso está íntegro dentro del mismo año no necesita partir el for
    for jHoraDelPaso := 0 to ceil( globs.HorasDelPaso - 1 ) do
      PHoraria_[jHoraDelPaso] :=
        demAnioBase[jHoraDelAnio_desde + jHoraDelPaso] * factCrecimiento;
  end
  else // globs.FechaInicioDelpaso.anio < globs.FechaFinDelPaso.anio
  begin // el paso comienza en un año y termina en el siguiente
    for jHoraDelPaso := 0 to (jHoraDelAnio_UltimaHoraDelAnio - jHoraDelAnio_desde) do
      PHoraria_[jHoraDelPaso] :=
        demAnioBase[jHoraDelAnio_desde + jHoraDelPaso] * factCrecimiento;

    ActualizarFactorYUltimaHoraDelAnio(globs.AnioInicioDelPaso + 1);

    for jHoraDelPaso := (jHoraDelAnio_UltimaHoraDelAnio - jHoraDelAnio_desde) +
      1 to (jHoraDelAnio_hasta - jHoraDelAnio_desde) do
      PHoraria_[jHoraDelPaso] := demAnioBase[jHoraDelPaso] * factCrecimiento;
  end;
  end;

  inherited prepararPaso_as;
end;

procedure TDemandaAnioBaseEIndices.Free;
begin
  fIniAnioBase.Free;
  if demAnioBase <> nil then
    SetLength(demAnioBase, 0);
  SetLength(GWh_anios, 0);
  ArchiDatos.Free;
  inherited Free;
end;

procedure TDemandaAnioBaseEIndices.InitFromFile;
var
  datos: TDatosHorariosDetallados;
  i: integer;
begin
  datos := TDatosHorariosDetallados.Create(self.ArchiDatos.archi);
  SetLength(demAnioBase, 366 * 24);
  datos.ReadBuff_horario(demAnioBase, datos.fechaPrimerDia);

  fIniAnioBase := TFecha.Create_Dt(datos.fechaPrimerDia);

  invSDem365 := 0;
  for i := 0 to 365 * 24 - 1 do
  begin
    invSDem365 := invSDem365 + demAnioBase[i];
  end;
  invSDem366 := invSDem365;
  for i := 365 * 24 to 366 * 24 - 1 do
    invSDem366 := invSDem366 + demAnioBase[i];

  invSDem365 := 1000 / invSDem365; //MWh --> GWh
  invSDem366 := 1000 / invSDem366;
  datos.Free;
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(TDemandaAnioBaseEIndices.ClassName, TDemandaAnioBaseEIndices);
end;

procedure AlFinal;
begin
end;

end.
