unit uHidroConEmbalse;

// Si setá definida, habilita el control de crecida.
{$DEFINE CONCONTROLCRECIDA}

// Prueba.. ... para imponer en los bordes del intervalo de discretización
// que multiplique por 2 (Dos, en el bode de menor cota) o por 0 (Cero, en el de
// mayor cota.
{xDEFINE DERIV_20}



// Durante la optmización usa el mismo método de derviación
// cotinua que durante la simlación. Si no está definido, usa
// un método de derivada en "estrellita" que puede ser más eficiente
{$DEFINE OPT_DERIV_CONTINUA}


// Chequea que la decremental sea superior a la incrementa, si no sucede
// las intercambia
{xDEFINE HIDRO_DERIVADA_MONOTONA}

// Si se define, luego de calcular las derivadas incremental y decremental
// pone ambas iguales al promedio.
{$DEFINE HIDRO_DERIVADA_SIMETRICA}

// Si se define, luego de calcular las derivadas incremental y decremental
// pone ambas iguales al maximo. Esta opción es incompatible con la anterior
{xDEFINE HIDRO_DERIVADA_MAX}




// Itera sobre la cota para cálculo del coeficiente energético
// si la cota varia sustancialmente durante el paso
{$DEFINE USAR_APRI}

(**
REVISIONES.

rch+20120426. v3.40
Agregué en el cálculo del máximo volumen exigible en las restricciones
de erogado mínimo que tenga en cuenta además del volúmen embalsado y los aportes, la
capacidad de extraer ese volúmen mediante turivinado+vertido. Tal como estaba no se hacía
ese control y en situaciones en que el lago se encuentra por debajo del vertedero
si se le pedía un erogado por encima del máximo turbinable no lo lograba.
Ahora, para determinar si puede cumplir con una condición de Erogado Mínimo,
Limita el volumen de la condición a no superar ni el volumen embalsado + Aportes - pérdidas del paso
y a no superar el máximo volumen erogable (turbinado + vertido). En esta versión se
agregó el control del volumen erogable.

h_actual = Cota vista para manejo del stock. Refleja siempre la cota al inicio del paso.
h_real = Cota física para cálculo del salto y del coeficiente energético. Si hay una fuente
       que fija la cota física se toma de ahí. Sino, es igual a h_actual si no hay iteraciones
       y si las hay refleja la cota media del paso para cálculo del coeficiente energético.

**)

interface

uses
  Classes,
  Math, SysUtils,
  xMatDefs, MatReal, uSimplex, ucosa, uGlobs, uFechas, uNodos,
  uRandDispos,
  uEstados, uodt_types,
  uFichasLPD,
  uGeneradores,
  uconstantesSimSEE,
  uFuentesAleatorias, uCosaConNombre, uevapUruguay,
  compol;

resourcestring
  rsHidroelectricaEmbalse = 'Hidroeléctrica con embalse';

const
  ro_Agua = 1000; // kg/m3 densidad del agua
  g_Gravitacion = 9.8; // N/kg constante gravitatoria




type
  THidroConEmbalse = class; // forward declaration

(* Para describir el embalse, damos el volumen máximo turbinable Vmax
que corresponde al volumen (en m3 ) que almacena el embalse entre las
cotas hmin y hmax.
Y el volumen almacenado entre hmin y hmed siendo hmed = (hmax+hmin)/2

Las cotas están medidas desde algún nivel (por ejemplo el del mar).
Para poder calcular el salto se suministra también la cota de aguas abajo
o de descarga de la turbina referida al mismo nivel. (hDescarga [m])
El nivel de descarga puede ser especificado dando su valor con la variable
hDescarga o mediante la referencia a otra central en cuyo caso se considera
que el nivel de descarga es el del lago de la otra central más el valor
hDescarga. En este caso la variable hDescarga se utiliza para tener en
cuenta cambios en la referencia de niven entre una central y otra.
Si todas las cotas están referidas al nivel del mar no hay cambios de
referencias y hDescarga se pondrá en cero.

Suponemos que el área del embalse crece linealmente con la cota
A(z) = a + b * z
V(z) = integral( A(z) )= a*z + 1/2 * b z^2

con z= (h-hmin)/(hmax-hmin)

Vmax= V(1) = a+ 0.5 b
Vhmed= V(0.5) = 0.5 a + 0.5^3 b

=> Vmax - 2*Vhmed= 0.5 b - 0.5^2 b =  0.5^2 b
=>  b = 4*(Vmax-2*Vhmed)
.:  a = Vmax - 0.5 b

V= ( 1/2*b/dhmax^2 * dh + a/dhmax ) dh =

Pérdida de Salto por Caudal Erogado.
====================================
El caudal erogado QErogado [m3/s] es calculado al final de cada paso
de simulación como la suma del caudal vertido más el caudal turbinado.
Dicho caudal puede afectar el nivel de aguas abajo reducciendo el salto
efectivo en la central. Para tener en cuenta este efecto se usan los
parámetros: caQE y cbQE calculados para modelar la pérdida de salto
de acuerdo a la siguiente ecuación:

dh(QE) = caQE* QE + cbQE* QE^2
*)
  TFichaHidroConEmbalse = class(TFichaLPD)
  public
    // Pagos a los generadores
    PagoPorDisponibilidad_USD_MWh: NReal; // [USD/MWh] Pago por Potencia
    PagoPorEnergia_USD_MWh: NReal; // [USD/MWh] Pago por Potencia


    // descripción de la geometria del lago
    PuntosCotaVolumen_h: TDAOfNReal;
    PuntosCotaVolumen_V: TDAOfNReal;

    // Si la fuente en lugar de caudal en m3/s es de escurrimiento
    // en mm/mes hay que especificar el área de la cuenca a considerar
    AreaCuenca_ha: NReal;

    // límites operativos
    hmax: NReal; //[m] cota maxima  de operación
    hmin: NReal; //[m] cota minima  de operación

    saltoMinimoOperativo: NReal; // [m] salto mínimo operativo de las turbinas

    centralesAguasArriba: TListaCentralesAguasArriba; //of TFichaCentralAguasArriba;
    hDescarga: NReal; //[m] cota de la descarga para cálculo del salto
    central_lagoDescarga: TGeneradorHidraulico;
    // Coeficientes de afectación del salto por caudal erogado
    // dh(QE) = caQE* QE + cbQE* QE^2
    caQE: NReal;
    cbQE: NReal;

    //La cota minima a partir de la cual se puede verter y la cota maxima a
    //partir de la cual el máximo vertimiento es constante
    cotaMV0, cotaMV1: NReal;
    //El caudal que se puede verter con la cota maxima.
    //QMV0 se omite porque siempre es 0
    QMV1: NReal;

    // caudal en m3/s mínimo a erorgar a la cota máxima de especificación
    // de la curva de erogado mínimo para protección de la presa.
    QE_ControlDeCrecidaAPleno: NReal;

    ren: NReal;       //[pu] rendimiento complexivo de turbina y generador
    Pmax_Gen: NReal;  //[MW] Potencia maxima hidraulica
    Qmax_Turb: NReal; //[m3/s]
    fDispo: NReal;    //[pu] factor de disponibilidad fortuita
    tRepHoras: NReal;

    // filtración = filtracion_Ca + filtracion_Cb * ( h- hmin )
    filtracion_Ca: NReal; //[m3/s]
    filtracion_Cb: NReal; //[m2/s]

    // Caudal de aportes por debajo del cual consideramos ambiente MUY SECO
    // para el cálculo de la evaporación
    QaMuySeco: NReal; //[m3/s]

    HayRestriccionEmaxPasoDeTiempo: boolean; // indica si se aplica la restricción
    EmaxPasoDeTiempo: NReal; // Energía maxima generable en un paso de tiempo

    HayRestriccionQTmin: boolean; // indica si se aplica la restricción
    QTmin: NReal; // caudal mínimo para asegurar navegabilidad
    ImponerQminPorPoste: boolean;

    (* rch080402 agrego esto para dar la opción de regular el uso de Salto y Palmar *)
    flg_controlCotaObjetivoInferior: boolean;
    flg_controlCotaObjetivoSuperior: boolean;
    hObjetivo: NReal; // [m] cota objetivo
    delta_cva_ParaControlDeCota: NReal; // USD/Hm3 aplicable para control de cota
    cv_USD_Hm3_ValorizadoManual: NReal;

    tomarCotaDeLaFuente: boolean;
    fuenteCota: TFuenteAleatoria;
    borneCota: string;
    numeroBorneCota: integer;

    cotaControlCrecida_Inicio: NReal;
    //Altura a la cual se activa el control de crecida en m
    cotaControlCrecida_Pleno: NReal;
    //a esta altura se tiene ErogadoMínimo = maximo del vertedero


    flg_CalcularEvaporacionDelLago: boolean;
    flg_CalcularFiltracionDelLago: boolean;

    constructor Create(capa: integer; fecha: TFecha; periodicidad: TPeriodicidad;
      hmin_, hmax_: NReal; PuntosCotaVolumen_h, PuntosCotaVolumen_V: TDAOfNReal;
      AreaCuenca_ha: NReal;
      hDescarga: NReal; central_lagoDescarga: TGeneradorHidraulico;
      centralesAguasArriba: TListaCentralesAguasArriba;
      caQE, cbQE, ren_, Pmax_Gen_, Qmax_Turb_: NReal; fDispo_: NReal;
      tRepHoras: NReal; filtracion_Ca, filtracion_Cb, QaMuySeco: NReal;
      cotaMV0, cotaMV1, QMV1, QE_ControlDeCrecidaAPleno: NReal;
      HayRestriccionEmaxPasoDeTiempo: boolean; EmaxPasoDeTiempo: NReal;
      HayRestriccionQTmin: boolean; QTmin: NReal; ImponerQminPorPoste: boolean;
      ControlarCotaPorDebajoDeObjetivo: boolean; ControlarCotaPorEnciaDeObjetivo: boolean;
      hObjetivo: NReal; Delta_cvaUSD_Hm3ParaControlDeCota: NReal;
      cv_USD_Hm3_ValorizadoManual: NReal; tomarCotaDeLaFuente: boolean;
      fuenteCota: TFuenteAleatoria; borneCota: string; saltoMinimoOperativo: NReal;
      cotaControlCrecida_Inicio: NReal; cotaControlCrecida_Pleno: NReal;
      flg_CalcularEvaporacionDelLago: boolean; flg_CalcularFiltracionDelLago: boolean;
      PagoPorDisponibilidad_USD_MWh: NReal; PagoPorEnergia_USD_MWh: NReal);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    function infoAd_: string; override;
    procedure Free; override;
  end;

  THidroConEmbalse = class(TGeneradorHidraulico)
  public
    pa: TFichaHidroConEmbalse;

    fuenteDeAportes: TFuenteAleatoria;
    nombreBorne: string;
    numeroBorne: integer;

    flg_FuenteDeEscurrimientos: boolean; //
      // Si es TRUE, la fuente suministra el escurrimiento en mm/mes
      // y para calcular el caudal hay que multiplicar por el área
      // y dividir por los segundos del mes.  (mm/mes) /1000 * Area[há]*100*100 /(730*3600)
      // si es false, la fuente
      // suministra el caudal medio en m3/s.


    //Variables de Estado
    NDisc: integer; // discretizacion del estado
    hini: NReal; // cota inicial



    //----------- variables auxiliares
    // Calcular con cambioFicha.
    polinomioCotaVolumen: TPoliR;
    V_hmin: NReal;
    V_hmax: NReal; //[Hm3] volumen máximo almacenable
    V_hmed: NReal; //[Hm3] volumen con la cota media

    // Factor Escurrimiento To Caudal
    fEscurrimientoToCaudal: NReal;

{$IFDEF CONCONTROLCRECIDA}
    V_ControlCrecida_Inicio: NReal; // [Hm3] al que se inicia el control de crecida
{$ENDIF}
    ro_g_ren: NReal; // ro_Agua * g_gravitacion * ren /(1E6*3600)
    dhmax: NReal; // hmax - hmin

    VTmin: NReal; // volumen turbinado mínimo en el paso para cumplir con QTmin
    QTmin_CalcVTminsobreHorasDelPaso: NReal;
    // caudal en Hm3/h para restricciones por poste.

    AplicarRestriccionDeErogadoMinimo: boolean;

    h_actual: NReal; // cota actual. Es la de toma.
    h_real: NReal;   //cota actual sin considerar la fuenteCota

    // variables auxiliares para iterar sobre la cota para cálculo del coeficiente energético
    apri_a, apri_b: NReal;
    apri_z1, apri_z2, apri_s, apri_delta_v, apri_z_aster: NReal;

    salto_actual: NReal; // salto actual efectivo
{$IFDEF ITERAR}
    QErogado_IterAnterior: NReal;
    // Usamos esta variable para almacenar el QErogado en la
    // iteración anterior. Esto se hace en leerResultadosDelPaso
    // antes de cargar el nuevo valor de QErogado.
{$ENDIF}
    QMaxVert: NReal;
    ce_MAXTURB0: NReal;


    (******** Esta es la variable de estado ******)
    X_VolumenEmbalsado: NReal; // Volumen Embalsado al inicio del paso
    Xs_VolumenEmbalsado: NReal; // Volumen Embalsado al final del paso

    PmaxParaQmax: NReal; // Pontecia máxima por máquina considerando Qmax y la cota
    NMaquinasDisponibles: integer;

    cv_agua_PenalidadPorRiesgoDePresa_USD_Hm3: NReal;
    // penalidad por exceder la cota máxima de operación.

    Pmax_Central: NReal; // [MW] Potencia Máxima disponible en la central
    Pmax_Gen: NReal; // [MW] Potencia máxima generable por turbina-generador
    Qmax: NReal; // [m3/s] Caudal máximo turbinable por turbina-generador

    dh_RedQE: NReal; // [m] Guarda la pérdida de salto por caulda erogado

    // variables auxiliares para la iteración por caudal erogado
    // Si está EnIteracion, el rango de potencias se fija en
    // el procedimiento NecesitoIterar para evitar oscilaciones
    P_Central_CotaSup: TDAOfNReal;


    PuedoGenerar: boolean;
    // indica si la Central está en condiciones de ofrecer potencia o no
    cv_USD_MWh: NReal; // costo variable de generación
    cv_agua_USD_Hm3_Dec: NReal; // valor del Agua expresado en USD/MWh

    Qevap: NReal; // caudal perdido por evaporacion
    Qfilt: NReal; // caudal perdido por filtración
    VPerdidas: NReal; //( QEvap+QFilt )* dt

    // variables auxiliares para las funciones cota-volumen y su inversa
    EsConico: boolean;
    embalse_cb, embalse_ca: NReal;
    embalse_c1, embalse_c2, embalse_c3_: NReal;
    embalse_c4_, embalse_c5_: NReal;

    flg_ModificandoQEmaxParaEstabilidad: boolean;
    flg_RecortandoAportesParaEstabilidad: boolean;

{$IFDEF ITERAR}
    //Coeficientes de amortiguamiento, el otro es 1 - alfa
    alfa: NReal;
{$ENDIF}

    flg_ValorizadoManual: boolean;


    Vs_SinErogado: NReal;
{$IFDEF CONCONTROLCRECIDA}
    // varialbes auxiliares de prepararPaso_ps
    hs_SinErogado: NReal;
    V_ErogadoMinimo_Ctrl_Crecida: NReal;
    Q_ErogadoMinimo_Ctrl_Crecida: NReal;
{$ENDIF}

    // auxiliar para preparar_ps
    auxParaCalculos_cv: NReal;
    V_EmbalsadoInicioVertedero: NReal;
    V_EmbalsadoFinalVertedero: NReal;
    Delta_V_Embalsado_Vertedero: NReal;

    V_EmbalsadoInicioProteccion: NReal;
    V_EmbalsadoProteccionAPleno: NReal;
    Delta_V_Embalsado_Proteccion: NReal;
    r_QE_Proteccion: NReal;

    constructor Create(capa: integer; nombre: string; nacimiento, muerte: TFecha;
      lpdUnidades, lpd: TFichasLPD; nodo: TNodo; hIni: NReal;
      NDisc_: integer;
      fuenteDeAportes: TFuenteAleatoria; nombreBorne: string;
      flg_FuenteDeEscurrimientos: boolean;
      flg_ValorizadoManual: boolean; TonCO2xMWh: NReal;
      LowCostMustRun, CleanDevelopmentMechanism: boolean);

    constructor Create_ReadFromText(f: TArchiTexto); override;
    procedure WriteToText(f: TArchiTexto); override;
    procedure PrepararMemoria(globs: TGlobs); override;
    procedure RegistrarParametrosDinamicos; override;

    function PotenciaFirme: NReal; override;

    function InfoAd_: string; override;
    class function DescClase: string; override;

    //      procedure Sim_Inicio; override;
    procedure Sim_Cronica_Inicio; override;

    procedure SorteosDelPaso(sortear: boolean); override;
    procedure PrepararPaso_ps; override;

    // Procedimiento auxiliar usado en PrepararPaso_ps
    procedure calcularDerivadas_USD_Hm3_SegunEstadoActual(
      var cv_agua_USD_Hm3_Inc, cv_agua_USD_Hm3_Dec: NReal);


    function ivarVertimientosMiosYAguasAbajo: TDAofNInt;

    procedure opt_nvers(var ivar, ivae, ires: integer); override;
    procedure opt_cargue(s: TSimplex); override;

    procedure EvolucionarEstado; override;


    procedure ResolverEncadenamientos(s: TSimplex); override;
    procedure opt_fijarRestriccionesDeCaja(s: TSimplex); override;
    procedure opt_leerSolucion(s: TSimplex); override;

{$IFDEF ITERAR}
    function opt_necesitoIterar(kIteracion: integer;
      var errRelativo: NReal): boolean;
      override;
{$ENDIF}

{$IFDEF SPXCONLOG}
    procedure spx_NombrarVariables(s: TSimplex); override;
{$ENDIF}
    function getNombreVar(ivar: integer; var nombre: string): boolean; override;
    function getNombreRes(ires: integer; var nombre: string): boolean; override;

    procedure PosicionarseEnEstrellita; override;
    procedure ActualizarEstadoGlobal; override;
    procedure AcumAux1(peso: NReal); override;
    procedure SetAux1; override;

    procedure optx_nvxs(var ixr, ixd, iauxr, iauxd: integer); override;
    procedure optx_RegistrarVariablesDeEstado(adminEstados: TAdminEstados);
      override;

    procedure Free; override;

    // geometría del embalse
    function CotaToVolumen(h: NReal): NReal;
    function VolumenToCota(Vol: NReal): NReal;

    // resultado m2
    function CotaToSuperficie_(h: NReal): NReal;

    procedure CambioFichaPD; override;

    procedure PubliVars; override;

    function CotaAguasArriba: NReal; override;
    // Calcula la cota de la descarga para poder calcular el Salto
    function CotaDescarga: NReal; override;
    // Caclula la reducción del salto causada por el caudal Erogado
    function ReduccionDeCotaPorCaudalErogado(QErogado: NReal): NReal; override;
    function centralLagoDescarga: TGeneradorHidraulico; override;

    // imprime tabla de valores del agua en USD/Hm3
    procedure opt_PrintResultados_Encab(var fsal: textfile); override;
    procedure opt_PrintResultados(var fsal: textfile); override;

    procedure dump_Variables(var f: TextFile; charIndentacion: char); override;
    procedure sim_FinCronicaPrintEstadoFinal(var fsal: textfile); override;

    function vertimientoPorPoste: boolean; override;

    function get_pa_FD( kTipoUnidad: integer ): NReal; override;
    function get_pa_TMR( kTipoUnidad: integer ): NReal; override;

  end;

procedure THidroConEmbalse_cambioFichaPD(Actor: TCosa);

procedure AlInicio;
procedure AlFinal;

implementation

uses uActores;

{$IFDEF DBG_HIDRO_CON_EMBALSE}
var
   flog: textfile;
   flg_flog_open: boolean;
{$ENDIF}

constructor THidroConEmbalse.Create(capa: integer; nombre: string;
  nacimiento, muerte: TFecha; lpdUnidades, lpd: TFichasLPD; nodo: TNodo;
  hIni: NReal; NDisc_: integer; fuenteDeAportes: TFuenteAleatoria;
  nombreBorne: string;
  flg_FuenteDeEscurrimientos: boolean;
  flg_ValorizadoManual: boolean; TonCO2xMWh: NReal;
  LowCostMustRun, CleanDevelopmentMechanism: boolean);
begin
  inherited Create(capa, nombre, nacimiento, muerte, lpdUnidades, nodo,
    TonCO2xMWh, LowCostMustRun, CleanDevelopmentMechanism);
  flg_ModificandoQEmaxParaEstabilidad := False;
  flg_RecortandoAportesParaEstabilidad := False;

  Self.lpd := lpd;
  self.hini := hIni;
  self.NDisc := NDisc_;
  self.fuenteDeAportes := fuenteDeAportes;
  self.nombreBorne := nombreBorne;
  self.flg_FuenteDeEscurrimientos:= flg_FuenteDeEscurrimientos;
  self.flg_ValorizadoManual := flg_ValorizadoManual;
end;

constructor THidroConEmbalse.Create_ReadFromText(f: TArchiTexto);
begin
  inherited Create_ReadFromText(f);
  flg_FuenteDeEscurrimientos:= false;

  f.IniciarLecturaRetrasada;
  f.rd('hIni', hIni);
  f.rd('NDisc', NDisc);
  f.rd('lpd', TCosa(lpd));
  f.rdReferencia('fuenteDeAportes', TCosa(fuenteDeAportes), Self);
  f.rd('nombreBorne', nombreBorne);

  if f.Version > 104 then
    f.rd( 'flg_FuenteDeEscurrimientos', flg_FuenteDeEscurrimientos );

  f.rd('flg_ValorizadoManual', flg_ValorizadoManual);
  f.EjecutarLectura;
  flg_ModificandoQEmaxParaEstabilidad := False;
  flg_RecortandoAportesParaEstabilidad := False;

  pa := nil;
  nodo := nil;
  lpd.Propietario := self;
end;

procedure THidroConEmbalse.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('hIni', hini, CF_PRECISION, CF_DECIMALES);
  f.wr('NDisc', NDisc);
  f.wr('lpd', lpd);
  f.wrReferencia('fuenteDeAportes', fuenteDeAportes);
  f.wr('nombreBorne', nombreBorne);
  f.wr( 'flg_FuenteDeEscurrimientos', flg_FuenteDeEscurrimientos );
  f.wr('flg_ValorizadoManual', flg_ValorizadoManual);
end;

procedure THidroConEmbalse.prepararMemoria(globs: TGlobs);
begin
  inherited prepararMemoria(globs);
  setlength(P_Central_CotaSup, globs.NPostes);
  dh_RedQE := 0;

  polinomioCotaVolumen := TPoliR.Create_Init(2);
  numeroBorne := fuenteDeAportes.idBorne(nombreBorne);

  {$IFDEF DBG_HIDRO_CON_EMBALSE}
  if not flg_flog_open  then
  begin
    assignfile( flog, 'hidroemblog_'+  globs.EscenarioActivo.nombre + '.xlt' );
    rewrite( flog );
    writeln( flog, 'kEstrella'#9' kPaso'#9'X_Vol'#9'Xs_Vol'#9'QTurb'#9'cvaUSD/MWh' );
    flg_flog_open:= true;
  end;
  {$ENDIF}

end;

procedure THidroConEmbalse.registrarParametrosDinamicos;
var
  i: integer;
  ficha: TFichaHidroConEmbalse;
begin
  inherited registrarParametrosDinamicos;
  lpd.expandirFichas(globs);
  lpd.RegistrarFichasAActualizar(Self, globs.ActualizadorLPD, @pA, nil,
    THidroConEmbalse_cambioFichaPD);
  for i := 0 to lpd.Count - 1 do
  begin
    ficha := TFichaHidroConEmbalse(lpd[i]);
    if ficha.fuenteCota <> nil then
      ficha.numeroBorneCota := ficha.fuenteCota.IdBorne(ficha.borneCota);
  end;
end;

function THidroConEmbalse.infoAd_: string;
begin
  Result := '';
end;

function THidroConEmbalse.potenciaFirme: NReal;
begin
  Result := (paUnidades.nUnidades[0]) * pa.Pmax_Gen * pa.ren;
end;

class function THidroConEmbalse.DescClase: string;
begin
  Result := rsHidroelectricaEmbalse;
end;

procedure THidroConEmbalse.SorteosDelPaso(sortear: boolean);
begin
  if hayForzamientos or globs.ObligarDisponibilidad_1_ then
  begin
    NMaquinasDisponibles := paUnidades.nUnidades[0];
    Pmax_Gen := pA.Pmax_gen;
    Qmax := pA.Qmax_Turb;
  end
  else if (sortear) then
  begin
    NMaquinasDisponibles := Sorteos_RepRotUnidades;
    Pmax_Gen := pA.Pmax_gen;
    Qmax := pA.Qmax_Turb;
  end
  else
  begin
    NMaquinasDisponibles := paUnidades.nUnidades[0];
    Pmax_Gen := pA.Pmax_gen * pA.fdispo;
    Qmax := pA.Qmax_Turb * pA.fdispo;
  end;
end;

function THidroConEmbalse.ReduccionDeCotaPorCaudalErogado(QErogado: NReal): NReal;
begin
{$IFDEF REDUCCIONCOTAQE}
  Result := (pa.cbQE * QErogado + pa.caQE) * QErogado;
{$ELSE}
  Result := 0;
{$ENDIF}
end;

function THidroConEmbalse.centralLagoDescarga: TGeneradorHidraulico;
begin
  Result := pa.central_lagoDescarga;
end;

function THidroConEmbalse.cotaAguasArriba: NReal;
begin
  Result := h_actual;
end;

function THidroConEmbalse.CotaDescarga: NReal;
begin
  if pa.central_lagoDescarga = nil then
    Result := pa.hDescarga
  else
    Result := pa.central_LagoDescarga.cotaAguasArriba;
end;

procedure THidroConEmbalse.opt_PrintResultados_Encab(var fsal: textfile);
begin
  Write(fsal, #9, FloatToStrF(VolumenToCota(X_VolumenEmbalsado), ffgeneral, 6, 3));
end;

procedure THidroConEmbalse.opt_PrintResultados(var fsal: textfile);
begin
  Write(fsal, #9, FloatToStrF(cv_agua_USD_Hm3_Dec, ffgeneral, 6, 3));
end;

procedure THidroConEmbalse.dump_Variables(var f: TextFile; charIndentacion: char);
begin
  inherited dump_Variables(f, charIndentacion);
  writeln(f, charIndentacion, 'h_actual[m]= ', FloatToStrF(h_actual, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'salto_actual[m]= ',
    FloatToStrF(salto_actual, ffFixed, 10, 3));

  writeln(f, charIndentacion, 'QAportePropio[m^3/s]= ',
    FloatToStrF(QAportePropio, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'QFilt[m^3/s]= ', FloatToStrF(QFilt, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'QEvap[m^3/s]= ', FloatToStrF(QEvap, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'QMaxVert[m^3/s]= ',
    FloatToStrF(QMaxVert, ffFixed, 10, 3));

  writeln(f, charIndentacion, 'VolumenAlmacenado[hm^3]= ',
    FloatToStrF(Xs_VolumenEmbalsado, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'VMax[hm^3]= ', FloatToStrF(V_hmax, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'VAportePropio[hm^3]= ',
    FloatToStrF(VAportePropio, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'VPerdidas[hm^3]= ',
    FloatToStrF(VPerdidas, ffFixed, 10, 3));
(*
  writeln(f, charIndentacion, 'cv_agua_USD_MWh_Inc[USD/MWh]= ', FloatToStrF(cv_agua_USD_MWh_Inc, ffFixed, 10, 3));
  writeln(f, charIndentacion, 'cv_agua_USD_MWh_Dec[USD/MWh]= ', FloatToStrF(cv_agua_USD_MWh_Dec, ffFixed, 10, 3));
*)
  writeln(f);
end;

procedure THidroConEmbalse.sim_FinCronicaPrintEstadoFinal(var fsal: textfile);
begin
  writeln(fsal, Nombre + '.h_Real = ' + FloatToStr(self.h_real));
end;

procedure THidroConEmbalse.EvolucionarEstado;
begin
  X_VolumenEmbalsado := Xs_VolumenEmbalsado;
  h_actual := VolumenToCota(X_VolumenEmbalsado);
end;


procedure THidroConEmbalse.calcularDerivadas_USD_Hm3_SegunEstadoActual(
  var cv_agua_USD_Hm3_Inc, cv_agua_USD_Hm3_Dec: NReal);
var
  dCFdV_Inc, dCFdV_Dec: NReal;
  resCode: integer;
  aux: NReal;
begin
  // calculamos el valor del agua en USD/MWh
  if flg_ValorizadoManual then
  begin
    cv_agua_USD_Hm3_Inc := pa.cv_USD_Hm3_ValorizadoManual * globs.fActPaso;
    cv_agua_USD_Hm3_Dec := cv_agua_USD_Hm3_Inc;
  end
  else
  begin
{$IFDEF DERIV_20}
    if globs.EstadoDeLaSala = CES_OPTIMIZANDO then
      globs.CF.devxr_estrella_20(ixr, globs.kPaso_ + 1, dCFdV_Inc, dCFdV_Dec)
    else
      globs.CF.devxr_continuo_20(ixr, globs.kPaso_ + 1, dCFdV_Inc, dCFdV_Dec);
{$ELSE}
    if globs.EstadoDeLaSala = CES_OPTIMIZANDO then
    begin
{$IFDEF OPT_DERIV_CONTINUA}
      globs.CF.devxr_continuo(ixr, globs.kPaso_ + 1, dCFdV_Inc, dCFdV_Dec, resCode, aux);
{$ELSE}
      globs.CF.devxr_estrella(ixr, globs.kPaso_ + 1, dCFdV_Inc, dCFdV_Dec, resCode);
{$ENDIF}
    end
    else
    begin
      globs.CF.devxr_continuo(ixr, globs.kPaso_ + 1, dCFdV_Inc,
        dCFdV_Dec, resCode, aux);
    end;
{$ENDIF}
    cv_agua_USD_Hm3_Inc := -dCFdV_Inc * globs.fActPaso;
    cv_agua_USD_Hm3_Dec := -dCFdV_Dec * globs.fActPaso;
  end;



{$IFDEF HIDRO_DERIVADA_MONOTONA}
  if cv_agua_USD_Hm3_Dec < cv_agua_USD_Hm3_Inc then
     vswap( cv_agua_USD_Hm3_Inc, cv_agua_USD_Hm3_Dec );
{$ENDIF}

{$IFDEF HIDRO_DERIVADA_SIMETRICA}
  cv_agua_USD_Hm3_Dec := (cv_agua_USD_Hm3_Dec + cv_agua_USD_Hm3_Inc) / 2;
  cv_agua_USD_Hm3_Inc := cv_agua_USD_Hm3_Dec;
{$ENDIF}

{$IFDEF HIDRO_DERIVADA_MAX}
  cv_agua_USD_Hm3_Dec := max(cv_agua_USD_Hm3_Dec, cv_agua_USD_Hm3_Inc);
  cv_agua_USD_Hm3_Inc := cv_agua_USD_Hm3_Dec;
{$ENDIF}

  (**** AGREGADO PARA CONTROL DE COTA OBJETIVO DURANTE LA SIMULACION****)
  if globs.EstadoDeLaSala = CES_SIMULANDO then
  begin
    if pa.flg_controlCotaObjetivoInferior and (h_actual < pa.hObjetivo) then
    begin
      cv_agua_USD_Hm3_Dec := cv_agua_USD_Hm3_Dec + pa.delta_cva_ParaControlDeCota;
      cv_agua_USD_Hm3_Inc := cv_agua_USD_Hm3_Inc + pa.delta_cva_ParaControlDeCota;
    end;

    if pa.flg_controlCotaObjetivoSuperior and (h_actual > pa.hObjetivo) then
    begin
      cv_agua_USD_Hm3_Dec := cv_agua_USD_Hm3_Dec - pa.delta_cva_ParaControlDeCota;
      cv_agua_USD_Hm3_Inc := cv_agua_USD_Hm3_Inc - pa.delta_cva_ParaControlDeCota;
    end;
  end;

{$IFDEF COSTOSHIDROPOSITIVOS}
  if cv_agua_USD_Hm3_Inc < 1E-2 then
    cv_agua_USD_Hm3_Inc := 1e-2;

  if cv_agua_USD_Hm3_Dec < 1E-2 then
    cv_agua_USD_Hm3_Dec := 1e-2;
{$ENDIF}

end;

procedure THidroConEmbalse.prepararPaso_ps;
var
  ce_MAXTURB: NReal;
  VMaxErogable_Central: NReal;
  QMax_Erogable: NReal;
  QMaxTurbinable_Central: NReal;
  V_sobre_nivel_de_riesgo: NReal;

  {$IFDEF ITERAR}
  alfa, beta: NReal;
  k: integer;
  {$ENDIF}

  {$IFDEF USAR_APRI}
  apri_i1, apri_i2: NReal;
  {$ENDIF}
begin


  // supongo que puedo generar, luego si el salto no es suficiente
  // o si todas las unidades están indisponibles lo bajo a false
  PuedoGenerar := True;

  {$IFDEF ITERAR}
  if globs.cntIteracionesDelPaso = 1 then
  begin
  {$ENDIF}

    if not self.pa.tomarCotaDeLaFuente then
      h_real := h_actual
    else
      h_real := pa.fuenteCota.Bornera[pa.numeroBorneCota];

    calcularDerivadas_USD_Hm3_SegunEstadoActual(
      cv_agua_USD_Hm3_Inc, cv_agua_USD_Hm3_Dec);

  {$IFDEF ITERAR}
  end;

  alfa := 0.3;
  beta := 1 - alfa;

  QErogado_ := alfa * QErogado_IterAnterior + beta * QErogado_;
  QErogado_IterAnterior := QErogado_;

  QAportePropio := Self.fuenteDeAportes.bornera[Self.numeroBorne];
  if flg_FuenteDeEscurrimientos then
    QAportePropio:= QAportePropio * fEscurrimientoToCaudal;
(*
  Assert(QAportePropio >= 0, 'El generador hidráulico con embalse ' +
    nombre + ' tiene caudal de aportes propios negativo. Fuente: ' +
    fuenteDeAportes.nombre + ', Borne: ' + nombreBorne + FloatToStr(QAportePropio));
  *)
  QAportePropio:= max( 0, QAportePropio );



  if pa.flg_CalcularFiltracionDelLago then
    QFilt := pa.filtracion_Ca + pa.filtracion_Cb * (h_real - pa.hmin)
  else
    QFilt := 0.0;

  if pa.flg_CalcularEvaporacionDelLago then
    QEvap := CotaToSuperficie_(h_real) * CoeficienteDeEvaporacion_mps(
      globs.MesInicioDelPaso, QAportePropio, pa.QaMuySeco)
  else
    QEvap := 0.0;


  VAportePropio := QAportePropio * globs.SegundosDelPaso / 1.0E6;
  VPerdidas := (QFilt + QEvap) * globs.SegundosDelPaso / 1.0E6;

  // acotamos a que no pueda perderse más que lo que hay
  VPerdidas := max(0, min(VPerdidas, X_VolumenEmbalsado - V_hmin - (10 / 1.0E6)));


  {$IFDEF USAR_APRI}
  // en la primer pasada no hay nada que calcular.
  if (globs.cntIteracionesDelPaso > 1) and not self.pa.tomarCotaDeLaFuente then
  begin
    h_real := h_actual;

    apri_delta_v := (Xs_VolumenEmbalsado - X_VolumenEmbalsado) * 1.0e6;
    if abs(apri_delta_v) > 0.01 then
    begin
      apri_z1 := h_real - pa.hmin;
      apri_z2 := VolumenToCota(Xs_VolumenEmbalsado) - pa.hmin;
      apri_s := pa.hmin - CotaDescarga;
      if EsConico then
      begin
        apri_i2 := ((1 / 3.0 * apri_b * apri_z2 + (apri_a + apri_s * apri_b) / 2.0) *
          apri_z2 + apri_s * apri_a) * apri_z2;

        apri_i1 := ((1 / 3.0 * apri_b * apri_z1 + (apri_a + apri_s * apri_b) / 2.0) *
          apri_z1 + apri_s * apri_a) * apri_z1;

        apri_z_aster := (apri_i2 - apri_i1) / apri_delta_v;
      end
      else
        apri_z_aster := (((apri_a / 2.0) * apri_z2 + apri_s * apri_a) *
          apri_z2 - ((apri_a / 2.0) * apri_z1 + apri_s * apri_a) * apri_z1) /
          apri_delta_v;


      h_real := apri_z_aster + CotaDescarga;
    end;
  end;
  {$ENDIF}

  {$ENDIF}

  dh_RedQE := ReduccionDeCotaPorCaudalErogado(NMaquinasDisponibles * Qmax);
  salto_actual := max(0.1, h_real - CotaDescarga - dh_RedQE);
  ce_MAXTURB0 := salto_actual * ro_g_ren;

  //ro_g_ren= pA.ren * ro_Agua * g_gravitacion / (1.0E6 * 3600.0);

  dh_RedQE := ReduccionDeCotaPorCaudalErogado(QErogado_);
  salto_actual := max(0.1, h_real - CotaDescarga - dh_RedQE);

  if salto_actual < pa.saltoMinimoOperativo then
    self.PuedoGenerar := False; // no puedo mover las turbinas

  ce := salto_actual * ro_g_ren;

  ce_MAXTURB := min(ce_MAXTURB0, ce);

  auxParaCalculos_cv := 1 / (1.0E6 * ce);

  cv_USD_MWh := cv_agua_USD_Hm3_Dec * auxParaCalculos_cv;


  // en base al volumen actual calcula el vertimiento posible

  if X_VolumenEmbalsado < V_EmbalsadoInicioVertedero then
    QMaxVert := 0.0
  else
  begin
    if X_VolumenEmbalsado < V_EmbalsadoFinalVertedero then
      QMaxVert := pa.QMV1 * (X_VolumenEmbalsado - V_EmbalsadoInicioVertedero) /
        Delta_V_Embalsado_Vertedero
    else
      QMaxVert := pa.QMV1;
  end;


  PmaxParaQmax := Qmax * ce_MAXTURB * 3600;
  if NMaquinasDisponibles > 0 then
  begin
    Pmax_Central := NMaquinasDisponibles * min(Pmax_Gen, PMaxParaQmax);
    QMaxTurbinable_Central := PMax_Central / (ce * 3600.0);
  end
  else
  begin
    PMax_Central := 0;
    QMaxTurbinable_Central := 0;
  end;

  if PMax_Central <= 0.01 then
    PuedoGenerar := False;


  {$IFDEF ITERAR}
  if globs.cntIteracionesDelPaso = 1 then
  begin
    for k := 0 to globs.NPostes - 1 do
      P_Central_CotaSup[k] := Pmax_Central;
  end
  else
  begin
    Qmaxturbinable_Central := 0;
    for k := 0 to globs.NPostes - 1 do
    begin
      P_Central_CotaSup[k] := (P_Central_CotaSup[k] + P[k]) / 2.0;
      Qmaxturbinable_Central :=
        Qmaxturbinable_Central + P_Central_CotaSup[k] * globs.DurPos[k];
    end;
    QMaxTurbinable_Central :=
      QMaxTurbinable_Central / (ce * globs.SegundosDelPaso );
  end;
  {$ENDIF}


  QMax_Erogable := QMaxTurbinable_Central + QMaxVert;
  VMaxErogable_Central := QMax_Erogable * globs.SegundosDelPaso * (0.95 / 1.0E6);

  if pA.HayRestriccionQTmin then
  begin
    VTmin := X_VolumenEmbalsado - V_hmin + VAportePropio - VPerdidas;
    VTmin := 0.95 * VTmin; // lo bajo un poco cumplo si realmente puedo
    VTmin := min(PA.QTmin * globs.SegundosDelPaso / 1.0E6, VTmin);

    VTmin := min(VMaxErogable_Central, VTmin);

    QTmin_CalcVTminsobreHorasDelPaso := VTMin * globs.invHorasDelPaso;
    AplicarRestriccionDeErogadoMinimo := VTmin > 1;
  end
  else
  begin
    VTmin := 0;
    QTmin_CalcVTminsobreHorasDelPaso := 0;
    AplicarRestriccionDeErogadoMinimo := False;
  end;

  // no dejo desfondar el lago.
  Vs_SinErogado := max(X_VolumenEmbalsado + VAportePropio - VPerdidas, V_hmin);

{$IFDEF CONCONTROLCRECIDA}
  // ahora chequeamos si X_VolumenEmbalsado+VAportePropio - VPerdias > V(cotamaxop)
  // si eso pasa, imponemos un ErogadoMinimo que crece linealmente
  // dese 0 a la cota del maximo operativo hasta el Máximo del vertedero
  // para una coto 2m por arriba del máximo operativo.
  // En el futuro tenemos que dejar agregar la curva de erogado mínimo
  V_sobre_nivel_de_riesgo := Vs_SinErogado - V_ControlCrecida_Inicio;
  if (V_sobre_nivel_de_riesgo > 0) then
  begin
    Q_ErogadoMinimo_Ctrl_Crecida := V_sobre_nivel_de_riesgo * r_QE_Proteccion;

    // si el erogado mínimo es superior al Máximo Erogable lo acotamos
    if Q_ErogadoMinimo_Ctrl_Crecida > QMax_Erogable * 0.9 then
      Q_ErogadoMinimo_Ctrl_Crecida := QMax_Erogable * 0.9;

    V_ErogadoMinimo_Ctrl_Crecida :=
      Q_ErogadoMinimo_Ctrl_Crecida * globs.SegundosDelPaso / 1.0E6;

    // Imponemos el erogado como el máximo entre el requerido por l econtrol de crecida
    // y el requerido por otras razones
    if V_ErogadoMinimo_Ctrl_Crecida > VTMin then
    begin
      VTMin := V_ErogadoMinimo_Ctrl_Crecida;
      VTmin := min(VMaxErogable_Central, VTmin);
      QTmin_CalcVTminsobreHorasDelPaso := VTMin * globs.invHorasDelPaso;
      AplicarRestriccionDeErogadoMinimo := VTmin > 1;
    end;
  end;
{$ENDIF}

end;

procedure THidroConEmbalse.opt_nvers(var ivar, ivae, ires: integer);
begin

  self.ivar := ivar;

  if pa.ImponerQminPorPoste then
    ivar := ivar + globs.NPostes + globs.NPostes // PGen por poste y VVertimiento
  else
    ivar := ivar + globs.NPostes + 1; // PGen por poste y VVertimiento

  self.ires := ires; // 0 <= volumen_s
{$IFDEF CONCONTROLCRECIDA}
  ires := ires + 1;   // probando no usar <= Vmax
{$ELSE}
  // 0 <= volumen_s
  // Volumne_s <= Vmáx  ( dos restricciones )
  ires := ires + 2;
{$ENDIF}

  if (PuedoGenerar) and pA.HayRestriccionEmaxPasoDeTiempo then
    Inc(ires);

  if AplicarRestriccionDeErogadoMinimo then
    if pa.ImponerQminPorPoste then
      Inc(ires, globs.NPostes)
    else
      Inc(ires);
end;

procedure THidroConEmbalse.opt_cargue(s: TSimplex);
var
  ibaseres: integer;
  iposte: integer;
  jres: integer;
  invCEPorUnoESeis: NReal;
begin



{$IFDEF SPXCONLOG}
  spx_NombrarVariables(s);
{$ENDIF}
  invCEPorUnoESeis := 1 / (ce * 1.0E6);

  // **********************************
  // aporte a las restricciones de nodo como generador.
  ibaseres := nodo.ires;
  for iposte := 0 to globs.NPostes - 1 do
    s.pon_e(ibaseres + iposte, ivar + iposte, 1);
  // fin aporte a restricción de nodo
  //********************************


  // cargamos ivert con el índice en de columna donde empiezan los vertimientos
  ivert := ivar + globs.NPostes;

  // *****************************
  // Ahora agregamos a la función de UTILIDAD (-costo)

  // afectación del futuro por los turbinados
  for iposte := 0 to globs.NPostes - 1 do
    s.pon_e(s.nf, ivar + iposte, -cv_USD_MWh * globs.durpos[iposte] );

  // afectación del futuro por los vertimientos
  if pa.ImponerQminPorPoste then
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(s.nf, ivert + iposte, -cv_agua_USD_Hm3_Dec)
  else
    s.pon_e(s.nf, ivert, -cv_agua_USD_Hm3_Dec);

  // afectación del futuro por aportes - pérdidas (constantes del paso)
  s.acum_e(s.nf, s.nc, (cv_agua_USD_Hm3_Inc * VAportePropio -
    cv_agua_USD_Hm3_Dec * VPerdidas) ); // Aporte

  // fin de carga en la función de utilidad
  // *************************************++


  // *****************************************************
  // Ecuación de balance Vk+1 - V_hmin = Vk - V_hmin - Turbinados - Vertidos + Aportes >= 0
  jres := ires;

  // los turbinados
  if (PuedoGenerar) then
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(jres, ivar + iposte, -globs.durpos[iposte] * invCEPorUnoESeis);


  if pa.ImponerQminPorPoste then
  begin
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(jres, ivert + iposte, -1);
  end
  else
    s.pon_e(jres, ivert, -1);  // vertimiento


  // términos constantes
  s.pon_e(jres, s.nc, Vs_SinErogado - V_hmin );


  // fin de la ecuación de balance de volúmens
  //***********************************************************
{$IFNDEF CONCONTROLCRECIDA}
  // esta opción la dejo sólo para poder probar cómo daba antes, pero no debiera
  // usarse.
  // restricción de volumen <= Vmax
  Inc(jres);
  if (PuedoGenerar) then
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(jres, ivar + iposte, globs.durpos[iposte] * invCEPorUnoESeis);

  if pa.ImponerQminPorPoste then
  begin
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(jres, ivert + iposte, 1);
  end
  else
    s.pon_e(jres, ivert, 1);  // vertimiento

  //??? OJO el 10* es porque por ahora no quiero eliminar la restricción
  s.pon_e(jres, s.nc, Vmax - (X_VolumenEmbalsado + VAportePropio - VPerdidas));

  //??? ojo revisar si al agregar ImponerQminPorPoste no hay que tocar esto.
  //Si hay que cambiarlo, hay que permitir violar todos los postes, aunque con uno solo
  //alcanzaría
  s.permitirViolarBordeSupParaSatisfacerRestriccion(jres,
    ivarVertimientosMiosYAguasAbajo);
{$ENDIF}

  // Restricción de Energía máxima en el paso de tiempo
  if (PuedoGenerar) and pA.HayRestriccionEmaxPasoDeTiempo then
  begin
    Inc(jres);
    for iposte := 0 to globs.NPostes - 1 do
      s.pon_e(jres, ivar + iposte, -globs.durpos[iposte]);
    s.pon_e(jres, s.nc, pA.EmaxPasoDeTiempo);
  end;

  // Restricción de caudal erogado mínimo por navegabilidad
  // y Seguridad de la presa.
  if AplicarRestriccionDeErogadoMinimo then
  begin
    Inc(jres);
    if pa.ImponerQminPorPoste then
    begin
      for iposte := 0 to globs.NPostes - 1 do
      begin
        if (PuedoGenerar) then
        begin
          s.pon_e(jres, ivar + iposte, globs.durpos[iposte] * invCEPorUnoESeis);
        end;
        s.pon_e(jres, ivert + iposte, 1);  // vertimiento
        s.pon_e(jres, s.nc, -QTmin_CalcVTminsobreHorasDelPaso * globs.DurPos[iposte]);
        Inc(jres);
      end;
    end
    else
    begin
      if (PuedoGenerar) then
      begin
        for iposte := 0 to globs.NPostes - 1 do
          s.pon_e(jres, ivar + iposte, globs.durpos[iposte] * invCEPorUnoESeis);
      end;
      s.pon_e(jres, ivert, 1);  // vertimiento
      s.pon_e(jres, s.nc, -VTmin);
    end;
  end;

end;


function THidroConEmbalse.vertimientoPorPoste: boolean;
begin
  Result := pa.ImponerQminPorPoste;
end;




procedure THidroConEmbalse.ResolverEncadenamientos(s: TSimplex);
var
  iCentrales: integer;
  gh: TFichaCentralAguasArriba;
begin
  // Aportes de las centrales aguas arriba
  // sumamos los aportes de las centrales aguas arriba
  // sobre la restricción de vertimiento
  for iCentrales := 0 to pa.centralesAguasArriba.Count - 1 do
  begin
    gh := TFichaCentralAguasArriba(pa.centralesAguasArriba[iCentrales]);
    gh.Central.CargarAportesAFilaCentralAguasAbajo(
      ires, // fila en la que cargar los aportes
      gh.coef, // coeficiente de llegada por el que multiplicar los aportes
      self.cv_agua_USD_Hm3_Inc,
      // valor del agua a considerar en la central aguas abajo.
      s);
  end;
{$IFNDEF CONCONTROLCRECIDA}
  for iCentrales := 0 to pa.centralesAguasArriba.Count - 1 do
  begin
    gh.Central.CargarAportesAFilaCentralAguasAbajo(
      ires + 1, // fila en la que cargar los aportes
      -gh.coef, // coeficiente de llegada por el que multiplicar los aportes
      0,   // si había algo que cargar lo cargó la anterior.
      s);
  end;
{$ENDIF}
end;

procedure THidroConEmbalse.opt_fijarRestriccionesDeCaja(s: TSimplex);
var
  iposte: integer;
  qmax_Hm3PorHora: double;
begin
  // Le fijamos como cota máxima PMax a la potencia en todos los postes
  if PuedoGenerar then
  begin
    {$IFDEF ITERAR}
    if globs.cntIteracionesDelPaso > 1 then
    begin
      if (not hayForzamientos) then
      begin
        for iposte := 0 to globs.NPostes - 1 do
          s.cota_sup_set(ivar + iposte, P_Central_CotaSup[iposte]);
      end
      else
      begin
        for iposte := 0 to globs.NPostes - 1 do
          s.FijarVariable(ivar + iposte, paForzamiento.P[0]);
      end;
    end
    else
    {$ENDIF}
    if (not hayForzamientos) then
    begin
      for iposte := 0 to globs.NPostes - 1 do
        s.cota_sup_set(ivar + iposte, PMax_Central);
    end
    else
    begin
      for iposte := 0 to globs.NPostes - 1 do
        s.FijarVariable(ivar + iposte, paForzamiento.P[0]);
    end;
  end
  else
    for iposte := 0 to globs.NPostes - 1 do
      s.FijarVariable(ivar + iposte, 0);


  if pa.ImponerQminPorPoste then
  begin
    qmax_Hm3PorHora := QMaxVert * globs.SegundosDelPaso / 1.0E6 / globs.HorasDelPaso;
    for iposte := 0 to globs.NPostes - 1 do
      // atención 11/2/2010 le pongo *10 a la cota sup para que desborde si precisa.
      s.cota_sup_set(ivert + iposte, qmax_Hm3PorHora * globs.DurPos[iposte]);
  end
  else
    s.cota_sup_set(ivert, QMaxVert * globs.SegundosDelPaso / 1.0E6);

end;

procedure THidroConEmbalse.Sim_Cronica_Inicio;
begin
  inherited Sim_Cronica_Inicio;
  QErogado_ := 0;
{$IFDEF ITERAR}
  QErogado_IterAnterior := 0; // para que de igual al correrlo dos veces seguidas
{$ENDIF}
  VErogado := 0;
  QVertido := 0;
  VVertido := 0;
  QTurbinado := 0;
  VTurbinado := 0;
  EnergiaGenerada := 0;
  X_VolumenEmbalsado := CotaToVolumen(hini);
  // para que se copie al inicio del paso a X_VolumenEmbalsadoç
  Xs_VolumenEmbalsado := X_VolumenEmbalsado;
  h_actual := hini;
end;

procedure THidroConEmbalse.opt_leerSolucion(s: TSimplex);
var
  iposte: integer;
begin
  Ingreso_PorEnergia := 0;
  if pa.PagoPorDisponibilidad_USD_MWh <> 0 then
    Ingreso_PorDisponibilidad :=
      Pmax_Central * pa.PagoPorDisponibilidad_USD_MWh * globs.HorasDelPaso
  else
    Ingreso_PorDisponibilidad := 0;

  CostoDirectoDelPaso := Ingreso_PorDisponibilidad;

  if PuedoGenerar then
  begin
    // recuperamos los valores de Potencia despachada
    EnergiaGenerada := 0;
    for iposte := 0 to globs.NPostes - 1 do
    begin
      P[iposte] := s.xval(ivar + iposte);
      EnergiaGenerada := EnergiaGenerada + P[iposte] * globs.durpos[iposte];
    end;
    Ingreso_PorEnergia := EnergiaGenerada * pa.PagoPorEnergia_USD_MWh;
    CostoDirectoDelPaso := CostoDirectoDelPaso + Ingreso_PorEnergia;
  end
  else
  begin
    for iposte := 0 to globs.NPostes - 1 do
      P[iposte] := 0;
    EnergiaGenerada := 0;
  end;

  if pa.ImponerQminPorPoste then
  begin
    VVertido := 0;

    for iposte := 0 to globs.NPostes - 1 do
      VVertido := VVertido + s.xval(ivert + iposte);
  end
  else
    VVertido := s.xval(ivert);

  VTurbinado := EnergiaGenerada / (ce * 1.0E6);
  VErogado := VTurbinado + VVertido;

  QVertido := VVertido * (1.0E6 * globs.invSegundosDelPaso);
  QTurbinado := VTurbinado * (1.0E6 * globs.invSegundosDelPaso);
  QErogado_ := VErogado * (1.0E6 * globs.invSegundosDelPaso);

  Xs_VolumenEmbalsado := s.yval(ires) + V_hmin;

  {$IFDEF DBG_HIDRO_CON_EMBALSE}
  if flg_flog_open then
   writeln( flog, globs.CF.ordinalEstrellaActual, #9, globs.kPaso_, #9, X_VolumenEmbalsado,#9, Xs_VolumenEmbalsado,#9, QTurbinado,#9, cv_USD_MWh);
  {$ENDIF}

end;

{$IFDEF ITERAR}

function THidroConEmbalse.opt_NecesitoIterar(kIteracion: integer;
  var errRelativo: NReal): boolean;
var
  qm: NReal;
begin
  qm := max(QErogado_, QErogado_IterAnterior);
  if qm > 1 then
    errRelativo := abs(QErogado_ - QErogado_IterAnterior) / qm
  else
    errRelativo := 0;
  Result := errRelativo > 0.1;
end;



{$ENDIF}

{$IFDEF SPXCONLOG}
procedure THidroConEmbalse.spx_NombrarVariables(s: TSimplex);
var
  iposte, jres: integer;
begin
  for iposte := 0 to globs.NPostes - 1 do
    s.set_NombreVar(ivar + iposte, Nombre + '_P[MW]' + IntToStr(iposte + 1));
  if pa.ImponerQminPorPoste then
  begin
    for iposte := 0 to globs.NPostes - 1 do
      s.set_NombreVar(ivar + globs.NPostes + iposte,
        Nombre + '_Vertimiento[Hm3]' + IntToStr(iposte + 1));
  end
  else
    s.set_NombreVar(ivar + globs.NPostes, Nombre + '_Vertimiento[Hm3]');

  jres := ires;
  s.set_NombreRest(jres, Nombre + '_VOL>=0');

{$IFNDEF CONCONTROLCRECIDA}
  Inc(jres);
  s.set_NombreRest(jres, Nombre + '_VOL<=Vmax');
{$ENDIF}

  Inc(jres);
  if PuedoGenerar and pA.HayRestriccionEmaxPasoDeTiempo then
  begin
    s.set_NombreRest(jres, Nombre + '_Emax');
    Inc(jres);
  end;

  if AplicarRestriccionDeErogadoMinimo then
    if pa.ImponerQminPorPoste then
    begin
      for iposte := 0 to globs.NPostes - 1 do
        s.set_NombreRest(jres + iposte, Nombre + '_QTmin' + IntToStr(iposte + 1));
    end
    else
      s.set_NombreRest(jres, Nombre + '_QTmin');

end;

{$ENDIF}

function THidroConEmbalse.getNombreVar(ivar: integer; var nombre: string): boolean;
var
  iaux: integer;
begin
  if (ivar >= self.ivar) and (ivar < self.ivar + globs.NPostes) then
  begin
    nombre := self.Nombre + '_P[MW]' + IntToStr(ivar - self.ivar + 1);
    Result := True;
  end
  else
  if pa.ImponerQminPorPoste then
  begin
    iaux := ivar - ivert + 1;
    if (iaux >= 1) and (iaux <= globs.npostes) then
    begin
      nombre := self.Nombre + '_Vertimiento[Hm3]' + IntToStr(iaux);
      Result := True;
    end
    else
      Result := False;
  end
  else
  begin
    if (ivar = ivert) then
    begin
      nombre := self.Nombre + '_Vertimiento[Hm3]';
      Result := True;
    end
    else
      Result := False;
  end;
end;

function THidroConEmbalse.getNombreRes(ires: integer; var nombre: string): boolean;
var
  jres: integer;
  iposte: integer;
  ires_Vs_gt_0, ires_EMax, ires_QEMin_1: integer;
begin
  jres := self.ires;
  ires_Vs_gt_0 := jres;
  Inc(jres);

  if (PuedoGenerar and pa.HayRestriccionEmaxPasoDeTiempo) then
  begin
    ires_EMax := jres;
    Inc(jres);
  end
  else
    ires_Emax := -1;

  if self.AplicarRestriccionDeErogadoMinimo then
  begin
    ires_QEMin_1 := jres;
    Inc(jres);
  end
  else
    ires_QEMin_1 := -1;


  if (ires = ires_Vs_gt_0) then
  begin
    nombre := self.nombre + '_VOL>=0';
    Result := True;
    exit;
  end;


  if ires = ires_EMax then
  begin
    nombre := self.nombre + '_res-EMax';
    Result := True;
    exit;
  end;


  if self.AplicarRestriccionDeErogadoMinimo then
    if pa.ImponerQminPorPoste then
    begin
      iposte := (ires - ires_QEMin_1 + 1);
      if (iposte >= 0) and (iposte <= globs.NPostes) then
      begin
        nombre := self.nombre + '_res-QTMin_Poste' + IntToStr(iposte);
        Result := True;
        exit;
      end;
    end
    else
    begin
      if ires = ires_QEMin_1 then
      begin
        nombre := self.nombre + '_res-QTMin_Paso';
        Result := True;
        exit;
      end;
    end;

  Result := False;
end;

procedure THidroConEmbalse.PubliVars;
begin
  inherited PubliVars;
  PublicarVariableNR('QAportesP', '[m3/s]', 8, 1, QAportePropio, True);
  PublicarVariableNR('QTurbinado', '[m3/s]', 8, 1, QTurbinado, True);
  PublicarVariableNR('QVertido', '[m3/s]', 8, 1, QVertido, True);
  PublicarVariableNR('h_real', '[m]', 6, 2, h_real, True);
  PublicarVariableNR('h', '[m]', 6, 2, h_actual, True);
  PublicarVariableNR('dh_RedQE', '[m]', 6, 2, dh_RedQE, True);
  PublicarVariableNR('CV_aguaDec', '[USD/Hm3]', 6, 1, cv_agua_USD_Hm3_Dec, True);
  PublicarVariableNR('CV_aguaInc', '[USD/Hm3]', 6, 1, cv_agua_USD_Hm3_Inc, True);
  PublicarVariableNR('CV_USD_MWh', '[USD/MWh]', 6, 1, cv_USD_MWh, True);
  PublicarVariableNR('Vertimiento', '[Hm^3]', 15, 1, VVertido, False);
  PublicarVariableNR('Cota', '[m]', 6, 1, h_actual, False);
  PublicarVariableNR('Vol', '[Hm^3]', 15, 1, self.X_VolumenEmbalsado, False);
  PublicarVariableNI('NMaqsDisponibles', '-', NMaquinasDisponibles, False);
  //  PublicarVariableNR('QErogado_IterAnterior', '[m3/s]', 8, 1, QErogado_IterAnterior , False);
  PublicarVariableNR('QErogado', '[m3/s]', 8, 1, QErogado_, False);
  PublicarVariableNR('QEvaporación', '[m3/s]', 8, 1, QEvap, False);
  PublicarVariableNR('QFiltración', '[m3/s]', 8, 1, QFilt, False);
end;

function THidroConEmbalse.ivarVertimientosMiosYAguasAbajo: TDAofNInt;
var
  nCentrales: integer;
  iter: TGeneradorHidraulico;
  res: TDAofNInt;
begin
  nCentrales := 1; //yo

  iter := centralLagoDescarga;  // agregado
  while iter <> nil do
  begin
    nCentrales := nCentrales + 1;
    iter := iter.centralLagoDescarga;
  end;

  SetLength(res, nCentrales);
  nCentrales := 0;
  iter := self;
  while iter <> nil do
  begin
    res[nCentrales] := iter.ivert;
    nCentrales := nCentrales + 1;
    iter := iter.centralLagoDescarga;
  end;
  Result := res;
end;

procedure THidroConEmbalse.optx_nvxs(var ixr, ixd, iauxr, iauxd: integer);
begin
  if not flg_ValorizadoManual then
  begin
    self.ixr := ixr;
    ixr := ixr + 1;
  end;
  Self.iauxr := iauxr;
  iauxr := iauxr + 1;
end;

procedure THidroConEmbalse.PosicionarseEnEstrellita;
begin
  if not flg_ValorizadoManual then
    X_VolumenEmbalsado := globs.CF.xr[ixr];
  Xs_VolumenEmbalsado := X_VolumenEmbalsado; // es necesario
  h_actual := VolumenToCota(X_VolumenEmbalsado);
  Self.QErogado_ := globs.Auxs_r0[iauxr][globs.CF.ordinalEstrellaActual];
{$IFDEF ITERAR}
  //Para que tenga el mismo estado de iteraciones al posicionarse en la estrella y
  //de igual en la optimización monohilo y multihilo
  Self.QErogado_IterAnterior := QErogado_;
{$ENDIF}
end;

procedure THidroConEmbalse.ActualizarEstadoGlobal;
begin
  if flg_ValorizadoManual then
    exit;
  globs.CF.xr[ixr] := X_VolumenEmbalsado;
end;

procedure THidroConEmbalse.AcumAux1(peso: NReal);
begin
  globs.Auxs_r1[iauxr][globs.CF.ordinalEstrellaActual] :=
    globs.Auxs_r1[iauxr][globs.CF.ordinalEstrellaActual] + Self.QErogado_ * peso;
end;

procedure THidroConEmbalse.SetAux1;
begin
  globs.Auxs_r1[iauxr][globs.CF.ordinalEstrellaActual] := Self.QErogado_;
end;

procedure THidroConEmbalse.optx_RegistrarVariablesDeEstado(
  adminEstados: TAdminEstados);
begin
  if flg_ValorizadoManual then
    exit;

  adminEstados.Registrar_Continua(
    ixr,
    V_hmin,
    V_hmax,
    ndisc,
    nombre + '_Vol', // nombre de la variable
    'Hm3' // unidades
    );
  adminEstados.RegistrarTraduccion_Continua(ixr,
    nombre + '_Cota',
    'm', self.VolumenToCota);
end;

procedure THidroConEmbalse.Free;
begin
  if polinomioCotaVolumen <> nil then
    polinomioCotaVolumen.Free;
  if lpd <> nil then
    lpd.Free;
  setlength(P_Central_CotaSup, 0);
  inherited Free;
end;

// h [m]  V [hm3]
function THidroConEmbalse.CotaToVolumen(h: NReal): NReal;
var
  dh: NReal;
  dV: NReal;
begin
  dh := h - pA.hmin;
  if EsConico then
    dV:= (embalse_c2 * dh + embalse_c1) * dh
  else
    dV := embalse_c1 * dh;
  result:= V_hmin + dV;
end;

function THidroConEmbalse.CotaToSuperficie_(h: NReal): NReal;
var
  dh: NReal;
begin
  dh := h - pa.hmin;
  if EsConico then
    Result := (2 * embalse_c2 * dh + embalse_c1) * 1.0E6
  else
    Result := embalse_c1 * 1.0E6;
end;

function THidroConEmbalse.VolumenToCota(Vol: NReal): NReal;
var
  dV: NReal;
begin
  dV:= Vol - V_hmin;

  if EsConico then
    Result := sqrt(embalse_c3_ * dV + embalse_c4_) + embalse_c5_
  else
    Result := dV / embalse_c1 + pA.hmin;
end;


function THidroConEmbalse.get_pa_FD( kTipoUnidad: integer ): NReal;
begin
  result:= pa.fDispo;
end;

function THidroConEmbalse.get_pa_TMR( kTipoUnidad: integer ): NReal;
begin
  result:= pa.tRepHoras;
end;

procedure THidroConEmbalse.cambioFichaPD;
var
  hmed: NReal;
  alfa_QE_Proteccion: NReal;
  dVmax, dVmed: NReal;
begin
  polinomioCotaVolumen.AproxTabla2D(2, self.pa.PuntosCotaVolumen_h,
    self.pa.PuntosCotaVolumen_V);


  if flg_FuenteDeEscurrimientos then
    fEscurrimientoToCaudal:= 1.0/1000 * pa.AreaCuenca_ha *100 *100 / (730 * 3600 );

  V_hmin:= polinomioCotaVolumen.valx(pA.hmin);
  V_hmax := polinomioCotaVolumen.valx(pA.hmax);
  hmed := (pa.hmax + pa.hmin) / 2;
  V_hmed := polinomioCotaVolumen.valx(hmed);

  dVmax:= V_hmax - V_hmin;
  dVmed:= V_hmed - V_hmin;

  ActualizarProbabilidadesReparacionYRotura_(pa.fDispo, pa.tRepHoras);
  ro_g_ren := pA.ren * ro_Agua * g_gravitacion / (1.0E6 * 3600.0);

  dhmax := (pA.hmax - pA.hmin);

  embalse_cb := 4.0 * (dVmax - 2.0 * dVmed);
  embalse_ca := dVmax - embalse_cb / 2.0;
  if (embalse_cb < 0) then
    raise Exception.Create('Revise los datos del embalse(' + Nombre +
      '), el área decrece al aumentar la cota.');
  if (embalse_ca < 0) then
    raise Exception.Create('Revise los datos del embalse(' + Nombre +
      '), el volumen parece anularse antes de la cota mínima.');

  embalse_c1 := embalse_ca / dhmax;


  {$IFDEF USAR_APRI}
  apri_a := embalse_ca * 1E6 / dhmax;
  apri_b := embalse_cb * 1E6 / sqr(dhmax);
  {$ENDIF}

  if EsCero(embalse_cb) then
  begin
    EsConico := False;
  end
  else
  begin
    EsConico := True;
    embalse_c2 := embalse_cb / 2.0 / sqr(dhmax);

    embalse_c3_ := 1 / embalse_c2;
    embalse_c4_ := sqr(embalse_c1 / (2 * embalse_c2));
    embalse_c5_ := pA.hmin - (embalse_c1 / (2 * embalse_c2));
(*  otra forma de calcular lo mismo
   embalse_c3_ := 2.0 / ( embalse_cb/ sqr( dhmax) );
   embalse_c4_ := sqr( ( embalse_ca / dhmax)  / ( embalse_cb / sqr(dhmax) ));
   embalse_c5_ := pa.hmin - ( ( embalse_ca / dhmax)  / ( embalse_cb / sqr(dhmax) ));
   *)
  end;

  {$IFDEF CONCONTROLCRECIDA}
  // inicializamos el volumen de inicio del control de crecida.
  self.V_ControlCrecida_Inicio := self.CotaToVolumen(pa.cotaControlCrecida_Inicio);
  {$ENDIF}

  // calculo Volúmenes embalsados al inicio y fin del vertedero para facilitar cálculos
  V_EmbalsadoInicioVertedero := CotaToVolumen(pa.cotaMV0);
  V_EmbalsadoFinalVertedero := CotaToVolumen(pa.cotaMV1);
  Delta_V_Embalsado_Vertedero := V_EmbalsadoFinalVertedero - V_EmbalsadoInicioVertedero;

  // Cálculo de la constante k_QE para la fórmular de Caudal erogado para
  // protección de la presa.
  V_EmbalsadoInicioProteccion := CotaToVolumen(pa.cotaControlCrecida_Inicio);
  V_EmbalsadoProteccionAPleno := CotaToVolumen(pa.cotaControlCrecida_Pleno);
  Delta_V_Embalsado_Proteccion :=
    V_EmbalsadoProteccionAPleno - V_EmbalsadoInicioProteccion;
  alfa_QE_Proteccion := pa.QE_ControlDeCrecidaAPleno /
    (Delta_V_Embalsado_Proteccion * 1e6); // [1/s]
  r_QE_Proteccion := 1.0E6 / (globs.SegundosDelPaso) *
    (1 - exp(-alfa_QE_Proteccion * globs.SegundosDelPaso));
end;

{*********************************
*Métodos de TFichaHidroConEmbalse*
*********************************}

constructor TFichaHidroConEmbalse.Create(capa: integer; fecha: TFecha;
  periodicidad: TPeriodicidad; hmin_, hmax_: NReal;
  PuntosCotaVolumen_h, PuntosCotaVolumen_V: TDAOfNReal;
  AreaCuenca_ha: NReal;
  hDescarga: NReal;
  central_lagoDescarga: TGeneradorHidraulico;
  centralesAguasArriba: TListaCentralesAguasArriba;
  caQE, cbQE, ren_, Pmax_Gen_, Qmax_Turb_: NReal; fDispo_: NReal;
  tRepHoras: NReal; filtracion_Ca, filtracion_Cb, QaMuySeco: NReal;
  cotaMV0, cotaMV1, QMV1, QE_ControlDeCrecidaAPleno: NReal;
  HayRestriccionEmaxPasoDeTiempo: boolean; EmaxPasoDeTiempo: NReal;
  HayRestriccionQTmin: boolean; QTmin: NReal; ImponerQminPorPoste: boolean;
  ControlarCotaPorDebajoDeObjetivo: boolean; ControlarCotaPorEnciaDeObjetivo: boolean;
  hObjetivo: NReal; Delta_cvaUSD_Hm3ParaControlDeCota: NReal;
  cv_USD_Hm3_ValorizadoManual: NReal; tomarCotaDeLaFuente: boolean;
  fuenteCota: TFuenteAleatoria; borneCota: string; saltoMinimoOperativo: NReal;
  cotaControlCrecida_Inicio: NReal; cotaControlCrecida_Pleno: NReal;
  flg_CalcularEvaporacionDelLago: boolean; flg_CalcularFiltracionDelLago: boolean;
  PagoPorDisponibilidad_USD_MWh: NReal; PagoPorEnergia_USD_MWh: NReal);
begin
  inherited Create(capa, fecha, periodicidad);
  self.hmin := hmin_;
  self.hmax := hmax_;

  self.PuntosCotaVolumen_h := PuntosCotaVolumen_h;
  self.PuntosCotaVolumen_V := PuntosCotaVolumen_V;

  self.AreaCuenca_ha:= AreaCuenca_ha;
  self.hDescarga := hDescarga;
  self.central_lagoDescarga := central_lagoDescarga;
  self.caQE := caQE;
  self.cbQE := cbQE;
  self.ren := ren_;
  self.Pmax_Gen := Pmax_Gen_;
  self.Qmax_Turb := Qmax_Turb_;
  self.fDispo := fDispo_;
  self.tRepHoras := tRepHoras;

  Self.filtracion_Ca := filtracion_Ca;
  Self.filtracion_Cb := filtracion_Cb;
  Self.QaMuySeco := QaMuySeco;

  self.cotaMV0 := cotaMV0;
  self.cotaMV1 := cotaMV1;
  self.QMV1 := QMV1;
  self.QE_ControlDeCrecidaAPleno := QE_ControlDeCrecidaAPleno;

  self.centralesAguasArriba := centralesAguasArriba;

  self.HayRestriccionEmaxPasoDeTiempo := HayRestriccionEmaxPasoDeTiempo;
  self.EmaxPasoDeTiempo := EmaxPasoDeTiempo;
  self.HayRestriccionQTmin := HayRestriccionQTmin;
  self.QTmin := QTmin;
  self.ImponerQminPorPoste := ImponerQminPorPoste;

  self.flg_controlCotaObjetivoInferior := ControlarCotaPorDebajoDeObjetivo;
  self.flg_controlCotaObjetivoSuperior := ControlarCotaPorEnciaDeObjetivo;
  self.hObjetivo := hObjetivo;
  self.delta_cva_ParaControlDeCota := Delta_cvaUSD_Hm3ParaControlDeCota;
  self.cv_USD_Hm3_ValorizadoManual := cv_USD_Hm3_ValorizadoManual;

  Self.tomarCotaDeLaFuente := tomarCotaDeLaFuente;
  self.fuenteCota := fuenteCota;
  self.borneCota := borneCota;

  self.saltoMinimoOperativo := saltoMinimoOperativo;

  self.cotaControlCrecida_Inicio := cotaControlCrecida_Inicio;
  self.cotaControlCrecida_Pleno := cotaControlCrecida_Pleno;

  self.flg_CalcularEvaporacionDelLago := flg_CalcularEvaporacionDelLago;
  self.flg_CalcularFiltracionDelLago := flg_CalcularFiltracionDelLago;

  self.PagoPorDisponibilidad_USD_MWh := PagoPorDisponibilidad_USD_MWh;
  self.PagoPorEnergia_USD_MWh := PagoPorEnergia_USD_MWh;

end;

procedure TFichaHidroConEmbalse.WriteToText(f: TArchiTexto);
begin
  inherited WriteToText(f);
  f.wr('hmin', hmin, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('hmax', hmax, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('PuntosCotaVolumen_h', PuntosCotaVolumen_h, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('PuntosCotaVolumen_V', PuntosCotaVolumen_V, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr( 'AreaCuenca_ha', AreaCuenca_ha );
  f.wr('hDescarga', hDescarga, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('caQE', caQE, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('cbQE', cbQE, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('ren', ren, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('Pmax_Gen', Pmax_Gen, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('Qmax_Turb', Qmax_Turb, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('fDispo', fDispo, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);
  f.wr('tRepHoras', tRepHoras, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALESPU);

  f.wr('filtracion_Ca', filtracion_Ca, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('filtracion_Cb', filtracion_Cb, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('QaMuySeco', QaMuySeco, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);

  f.wr('cotaMV0', cotaMV0, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('cotaMV1', cotaMV1, uconstantesSimSEE.CF_PRECISION,
    uconstantesSimSEE.CF_DECIMALES);
  f.wr('QMV1', QMV1, uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);

  f.wr('QE_ControlDeCrecidaAPleno', QE_ControlDeCrecidaAPleno,
    uconstantesSimSEE.CF_PRECISION, uconstantesSimSEE.CF_DECIMALES);


  f.wrReferencia('central_lagoDescarga', central_lagoDescarga);
  f.wr('centralesAguasArriba', centralesAguasArriba);

  f.wr('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
  f.wr('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
  f.wr('HayRestriccionQTmin', HayRestriccionQTmin);
  f.wr('QTmin', QTmin);
  f.wr('ImponerQminPorPoste', ImponerQminPorPoste);

  f.wr('hObjetivo', hObjetivo);
  f.wr('cva_delta', delta_cva_ParaControlDeCota);
  f.wr('flg_controlCotaObjetivoInferior', flg_controlCotaObjetivoInferior);
  f.wr('flg_controlCotaObjetivoSuperior', flg_controlCotaObjetivoSuperior);

  f.wr('cv_USD_MWh_ValorizadoManual', cv_USD_Hm3_ValorizadoManual);

  f.wr('tomarCotaDeLaFuente', tomarCotaDeLaFuente);
  f.wrReferencia('fuenteCota', fuenteCota);
  f.wr('borneCota', borneCota);

  f.wr('cotaControlCrecida_Inicio', cotaControlCrecida_Inicio);
  f.wr('cotaControlCrecida_Pleno', cotaControlCrecida_Pleno);
  f.wr('saltoMinimoOperativo', saltoMinimoOperativo);


  f.wr('flg_CalcularEvaporacionDelLago', flg_CalcularEvaporacionDelLago);
  f.wr('flg_CalcularFiltracionDelLago', flg_CalcularFiltracionDelLago);

  f.wr('PagoPorDisponibilidad_USD_MWh', PagoPorDisponibilidad_USD_MWh);
  f.wr('PagoPorEnergia_USD_MWh', PagoPorEnergia_USD_MWh);

end;

function TFichaHidroConEmbalse.InfoAd_: string;
begin
  Result := 'PMáxGen= ' + FloatToStrF(Pmax_Gen, ffGeneral, 10, 1) +
    ' MW, ' + 'QMáxTurb= ' + FloatToStrF(Qmax_Turb, ffGeneral, 10, 1) +
    ' Hm^3/s, ' + 'ren= ' + FloatToStrF(ren, ffGeneral, 10, 2) +
    ' p.u., ' + 'fDisp= ' + FloatToStrF(fDispo, ffGeneral, 10, 2) +
    ' p.u., ' + 'tRep= ' + FloatToStrF(tRepHoras, ffGeneral, 10, 1) +
    'h, ' + 'hMín= ' + FloatToStrF(hmin, ffGeneral, 10, 2) + ' m, ' +
    'hMáx= ' + FloatToStrF(hmax, ffGeneral, 10, 2) + ' m, ' +
    'hDescarga= ' + FloatToStrF(hDescarga, ffGeneral, 10, 2) + 'm, ' +
    'Vmáx= ' + FloatToStrF(PuntosCotaVolumen_V[high(PuntosCotaVolumen_V)],
    ffGeneral, 10, 1) + ' m^3';
end;



constructor TFichaHidroConEmbalse.Create_ReadFromText(f: TArchiTexto);
var
  Vmax, Vhmed: NReal;
  vdummy: NReal;
label
  lbl_finLectura;
begin
  AreaCuenca_ha:= 0.0;

  if f.Version < 13 then
  begin
    inherited Create_ReadFromText(f);
    f.IniciarLecturaRetrasada;
    f.rd('hmin', hmin);
    f.rd('hmax', hmax);
    f.rd('Vmax', Vmax);
    f.rd('Vhmed', Vhmed);
    f.rd('hDescarga', hDescarga);
    f.rd('caQE', caQE);
    f.rd('cbQE', cbQE);
    f.rd('ren', ren);
    f.rd('Pmax_Gen', Pmax_Gen);
    f.rd('Qmax_Turb', Qmax_Turb);
    f.rd('fDispo', fDispo);
    f.rd('filtracion_Ca', filtracion_Ca);
    f.rd('filtracion_Cb', filtracion_Cb);
    f.rd('QaMuySeco', QaMuySeco);
    f.rd('cotaMV0', cotaMV0);
    f.rd('cotaMV1', cotaMV1);
    f.rd('QMV1', QMV1);
    f.rdReferencia('central_lagoDescarga', TCosa(central_lagoDescarga), Self);
    f.rd('centralesAguasArriba', TCosa(centralesAguasArriba));
    f.rd('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
    f.rd('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
    f.rd('HayRestriccionQTmin', HayRestriccionQTmin);
    f.rd('QTmin', QTmin);
    f.rd('hObjetivo', hObjetivo);
    f.rd('cva_delta', delta_cva_ParaControlDeCota);
    f.rd('flg_controlCotaObjetivoInferior', flg_controlCotaObjetivoInferior);
    f.rd('flg_controlCotaObjetivoSuperior', flg_controlCotaObjetivoSuperior);
    f.rd('cv_USD_MWh_ValorizadoManual', cv_USD_Hm3_ValorizadoManual);
    f.EjecutarLectura;
    tomarCotaDeLaFuente := False;
    fuenteCota := nil;
    borneCota := '';
    Self.tRepHoras := 48;
    ImponerQminPorPoste := False;
    cotaControlCrecida_Inicio := hmax;
    cotaControlCrecida_Pleno := cotaControlCrecida_Inicio + 2.0;
    goto lbl_finLectura;
  end;

  if f.Version < 14 then
  begin
    inherited Create_ReadFromText(f);
    f.IniciarLecturaRetrasada;
    f.rd('hmin', hmin);
    f.rd('hmax', hmax);
    f.rd('Vmax', Vmax);
    f.rd('Vhmed', Vhmed);
    f.rd('hDescarga', hDescarga);
    f.rd('caQE', caQE);
    f.rd('cbQE', cbQE);
    f.rd('ren', ren);
    f.rd('Pmax_Gen', Pmax_Gen);
    f.rd('Qmax_Turb', Qmax_Turb);
    f.rd('fDispo', fDispo);
    f.rd('filtracion_Ca', filtracion_Ca);
    f.rd('filtracion_Cb', filtracion_Cb);
    f.rd('QaMuySeco', QaMuySeco);
    f.rd('cotaMV0', cotaMV0);
    f.rd('cotaMV1', cotaMV1);
    f.rd('QMV1', QMV1);
    f.rdReferencia('central_lagoDescarga', TCosa(central_lagoDescarga), Self);
    f.rd('centralesAguasArriba', TCosa(centralesAguasArriba));
    f.rd('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
    f.rd('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
    f.rd('HayRestriccionQTmin', HayRestriccionQTmin);
    f.rd('QTmin', QTmin);
    f.rd('hObjetivo', hObjetivo);
    f.rd('cva_delta', delta_cva_ParaControlDeCota);
    f.rd('flg_controlCotaObjetivoInferior', flg_controlCotaObjetivoInferior);
    f.rd('flg_controlCotaObjetivoSuperior', flg_controlCotaObjetivoSuperior);
    f.rd('cv_USD_MWh_ValorizadoManual', cv_USD_Hm3_ValorizadoManual);
    f.rd('tomarCotaDeLaFuente', tomarCotaDeLaFuente);
    f.rdReferencia('fuenteCota', TCosa(fuenteCota), self);
    f.rd('borneCota', borneCota);
    f.EjecutarLectura;
    Self.tRepHoras := 48;
    ImponerQminPorPoste := False;
    cotaControlCrecida_Inicio := hmax;
    cotaControlCrecida_Pleno := cotaControlCrecida_Inicio + 2.0;
    goto lbl_finLectura;
  end;

  if f.Version < 17 then
  begin
    inherited Create_ReadFromText(f);
    f.IniciarLecturaRetrasada;
    f.rd('hmin', hmin);
    f.rd('hmax', hmax);
    f.rd('Vmax', Vmax);
    f.rd('Vhmed', Vhmed);
    f.rd('hDescarga', hDescarga);
    f.rd('caQE', caQE);
    f.rd('cbQE', cbQE);
    f.rd('ren', ren);
    f.rd('Pmax_Gen', Pmax_Gen);
    f.rd('Qmax_Turb', Qmax_Turb);
    f.rd('fDispo', fDispo);
    f.rd('tRepHoras', tRepHoras);
    f.rd('filtracion_Ca', filtracion_Ca);
    f.rd('filtracion_Cb', filtracion_Cb);
    f.rd('QaMuySeco', QaMuySeco);
    f.rd('cotaMV0', cotaMV0);
    f.rd('cotaMV1', cotaMV1);
    f.rd('QMV1', QMV1);
    f.rdReferencia('central_lagoDescarga', TCosa(central_lagoDescarga), Self);
    f.rd('centralesAguasArriba', TCosa(centralesAguasArriba));
    f.rd('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
    f.rd('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
    f.rd('HayRestriccionQTmin', HayRestriccionQTmin);
    f.rd('QTmin', QTmin);
    f.rd('hObjetivo', hObjetivo);
    f.rd('cva_delta', delta_cva_ParaControlDeCota);
    f.rd('flg_controlCotaObjetivoInferior', flg_controlCotaObjetivoInferior);
    f.rd('flg_controlCotaObjetivoSuperior', flg_controlCotaObjetivoSuperior);
    f.rd('cv_USD_MWh_ValorizadoManual', cv_USD_Hm3_ValorizadoManual);
    f.rd('tomarCotaDeLaFuente', tomarCotaDeLaFuente);
    f.rdReferencia('fuenteCota', TCosa(fuenteCota), self);
    f.rd('borneCota', borneCota);
    f.EjecutarLectura;
    ImponerQminPorPoste := False;
    cotaControlCrecida_Inicio := hmax;
    cotaControlCrecida_Pleno := cotaControlCrecida_Inicio + 2.0;
    goto lbl_finLectura;
  end;

  if f.Version < 26 then
  begin
    inherited Create_ReadFromText(f);
    f.IniciarLecturaRetrasada;
    f.rd('hmin', hmin);
    f.rd('hmax', hmax);
    f.rd('Vmax', Vmax);
    f.rd('Vhmed', Vhmed);
    f.rd('hDescarga', hDescarga);
    f.rd('caQE', caQE);
    f.rd('cbQE', cbQE);
    f.rd('ren', ren);
    f.rd('Pmax_Gen', Pmax_Gen);
    f.rd('Qmax_Turb', Qmax_Turb);
    f.rd('fDispo', fDispo);
    f.rd('tRepHoras', tRepHoras);
    f.rd('filtracion_Ca', filtracion_Ca);
    f.rd('filtracion_Cb', filtracion_Cb);
    f.rd('QaMuySeco', QaMuySeco);
    f.rd('cotaMV0', cotaMV0);
    f.rd('cotaMV1', cotaMV1);
    f.rd('QMV1', QMV1);
    f.rdReferencia('central_lagoDescarga', TCosa(central_lagoDescarga), Self);
    f.rd('centralesAguasArriba', TCosa(centralesAguasArriba));
    f.rd('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
    f.rd('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
    f.rd('HayRestriccionQTmin', HayRestriccionQTmin);
    f.rd('QTmin', QTmin);
    f.rd('ImponerQminPorPoste', ImponerQminPorPoste);
    f.rd('hObjetivo', hObjetivo);
    f.rd('cva_delta', delta_cva_ParaControlDeCota);
    f.rd('flg_controlCotaObjetivoInferior', flg_controlCotaObjetivoInferior);
    f.rd('flg_controlCotaObjetivoSuperior', flg_controlCotaObjetivoSuperior);
    f.rd('cv_USD_MWh_ValorizadoManual', cv_USD_Hm3_ValorizadoManual);
    f.rd('tomarCotaDeLaFuente', tomarCotaDeLaFuente);
    f.rdReferencia('fuenteCota', TCosa(fuenteCota), self);
    f.rd('borneCota', borneCota);
    f.EjecutarLectura;
    cotaControlCrecida_Inicio := hmax;
    cotaControlCrecida_Pleno := cotaControlCrecida_Inicio + 2.0;
    goto lbl_finLectura;
  end;

  if f.Version < 28 then
  begin
    inherited Create_ReadFromText(f);
    f.IniciarLecturaRetrasada;
    f.rd('hmin', hmin);
    f.rd('hmax', hmax);
    f.rd('Vmax', Vmax);
    f.rd('Vhmed', Vhmed);
    f.rd('hDescarga', hDescarga);
    f.rd('caQE', caQE);
    f.rd('cbQE', cbQE);
    f.rd('ren', ren);
    f.rd('Pmax_Gen', Pmax_Gen);
    f.rd('Qmax_Turb', Qmax_Turb);
    f.rd('fDispo', fDispo);
    f.rd('tRepHoras', tRepHoras);
    f.rd('filtracion_Ca', filtracion_Ca);
    f.rd('filtracion_Cb', filtracion_Cb);
    f.rd('QaMuySeco', QaMuySeco);
    f.rd('cotaMV0', cotaMV0);
    f.rd('cotaMV1', cotaMV1);
    f.rd('QMV1', QMV1);
    f.rdReferencia('central_lagoDescarga', TCosa(central_lagoDescarga), Self);
    f.rd('centralesAguasArriba', TCosa(centralesAguasArriba));
    f.rd('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
    f.rd('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
    f.rd('HayRestriccionQTmin', HayRestriccionQTmin);
    f.rd('QTmin', QTmin);
    f.rd('ImponerQminPorPoste', ImponerQminPorPoste);
    f.rd('hObjetivo', hObjetivo);
    f.rd('cva_delta', delta_cva_ParaControlDeCota);
    f.rd('flg_controlCotaObjetivoInferior', flg_controlCotaObjetivoInferior);
    f.rd('flg_controlCotaObjetivoSuperior', flg_controlCotaObjetivoSuperior);
    f.rd('cv_USD_MWh_ValorizadoManual', cv_USD_Hm3_ValorizadoManual);
    f.rd('tomarCotaDeLaFuente', tomarCotaDeLaFuente);
    f.rdReferencia('fuenteCota', TCosa(fuenteCota), self);
    f.rd('borneCota', borneCota);
    f.rd('cotaActivacionControlCrecida', vdummy);
    f.rd('penalidadControlCrecida', vdummy);
    f.EjecutarLectura;
    cotaControlCrecida_Inicio := hmax;
    cotaControlCrecida_Pleno := cotaControlCrecida_Inicio + 2.0;
    goto lbl_finLectura;
  end;

  // ultima versión - no borrar begin -end
  begin
    inherited Create_ReadFromText(f);
    f.IniciarLecturaRetrasada;
    f.rd('hmin', hmin);
    f.rd('hmax', hmax);
    f.rd('PuntosCotaVolumen_h', PuntosCotaVolumen_h);
    f.rd('PuntosCotaVolumen_V', PuntosCotaVolumen_V);

    if f.version > 104 then
      f.rd( 'AreaCuenca_ha', AreaCuenca_ha );
    f.rd('hDescarga', hDescarga);
    f.rd('caQE', caQE);
    f.rd('cbQE', cbQE);
    f.rd('ren', ren);
    f.rd('Pmax_Gen', Pmax_Gen);
    f.rd('Qmax_Turb', Qmax_Turb);
    f.rd('fDispo', fDispo);
    f.rd('tRepHoras', tRepHoras);
    f.rd('filtracion_Ca', filtracion_Ca);
    f.rd('filtracion_Cb', filtracion_Cb);
    f.rd('QaMuySeco', QaMuySeco);
    f.rd('cotaMV0', cotaMV0);
    f.rd('cotaMV1', cotaMV1);
    f.rd('QMV1', QMV1);
    if f.Version > 84 then
      f.rd('QE_ControlDeCrecidaAPleno', QE_ControlDeCrecidaAPleno);

    f.rdReferencia('central_lagoDescarga', TCosa(central_lagoDescarga), Self);
    f.rd('centralesAguasArriba', TCosa(centralesAguasArriba));
    f.rd('HayRestriccionEmaxPasoDeTiempo', HayRestriccionEmaxPasoDeTiempo);
    f.rd('EmaxPasoDeTiempo', EmaxPasoDeTiempo);
    f.rd('HayRestriccionQTmin', HayRestriccionQTmin);
    f.rd('QTmin', QTmin);
    f.rd('ImponerQminPorPoste', ImponerQminPorPoste);
    f.rd('hObjetivo', hObjetivo);
    f.rd('cva_delta', delta_cva_ParaControlDeCota);
    f.rd('flg_controlCotaObjetivoInferior', flg_controlCotaObjetivoInferior);
    f.rd('flg_controlCotaObjetivoSuperior', flg_controlCotaObjetivoSuperior);
    f.rd('cv_USD_MWh_ValorizadoManual', cv_USD_Hm3_ValorizadoManual);
    f.rd('tomarCotaDeLaFuente', tomarCotaDeLaFuente);
    f.rdReferencia('fuenteCota', TCosa(fuenteCota), self);
    f.rd('borneCota', borneCota);

    if f.Version < 29 then
    begin
      f.rd('cotaActivacionControlCrecida', vdummy);
      f.rd('penalidadControlCrecida', vdummy);
    end
    else
    begin
      f.rd('cotaControlCrecida_Inicio', cotaControlCrecida_Inicio);
      f.rd('cotaControlCrecida_Pleno', cotaControlCrecida_Pleno);
    end;

    if f.Version < 29 then
    begin
      cotaControlCrecida_Inicio := hmax;
      cotaControlCrecida_Pleno := cotaControlCrecida_Inicio + 2.0;
    end;

    if f.Version >= 30 then
    begin
      f.rd('saltoMinimoOperativo', saltoMinimoOperativo);
    end;
    if f.Version < 83 then
    begin
      flg_CalcularEvaporacionDelLago := True;
      flg_CalcularFiltracionDelLago := True;
    end
    else
    begin
      f.rd('flg_CalcularEvaporacionDelLago', flg_CalcularEvaporacionDelLago);
      f.rd('flg_CalcularFiltracionDelLago', flg_CalcularFiltracionDelLago);
    end;

    if f.Version < 98 then
    begin
      PagoPorDisponibilidad_USD_MWh := 0.0;
      PagoPorEnergia_USD_MWh := 0.0;
    end
    else
    begin
      f.rd('PagoPorDisponibilidad_USD_MWh', PagoPorDisponibilidad_USD_MWh);
      f.rd('PagoPorEnergia_USD_MWh', PagoPorEnergia_USD_MWh);
    end;

    f.EjecutarLectura;

    goto lbl_finLectura;
  end;

  lbl_finLectura:

    if f.Version < 28 then
    begin
      // calculo la tabla PuntosCotaVolumen a partir de los datos Vmax y Vhmed
      setlength(PuntosCotaVolumen_h, 3);
      setlength(PuntosCotaVolumen_V, 3);
      PuntosCotaVolumen_h[0] := self.hmin;
      PuntosCotaVolumen_h[1] := (self.hmin + self.hmax) / 2.0;
      PuntosCotaVolumen_h[2] := self.hmax;
      PuntosCotaVolumen_V[0] := 0;
      PuntosCotaVolumen_V[1] := Vhmed;
      PuntosCotaVolumen_V[2] := Vmax;
    end;

  if f.Version < 30 then
  begin
    saltoMinimoOperativo := 0.1;
  end;

  if f.Version < 85 then
    QE_ControlDeCrecidaAPleno := QMV1;

end;

procedure TFichaHidroConEmbalse.Free;
begin
  SetLength(PuntosCotaVolumen_h, 0);
  SetLength(PuntosCotaVolumen_V, 0);
  if centralesAguasArriba <> nil then
    centralesAguasArriba.Free;

  {$IFDEF DBG_HIDRO_CON_EMBALSE}
  if flg_flog_open then
  begin
   closefile( flog );
   flg_flog_open:= false;
  end;
  {$ENDIF}

  inherited Free;
end;

procedure THidroConEmbalse_cambioFichaPD(Actor: TCosa);
begin

  (Actor as THidroConEmbalse).cambioFichaPD;
end;

procedure AlInicio;
begin
  registrarClaseDeCosa(THidroConEmbalse.ClassName, THidroConEmbalse);
  registrarClaseDeCosa(TFichaHidroConEmbalse.ClassName, TFichaHidroConEmbalse);
end;

procedure AlFinal;
begin
end;

begin
{$IFDEF DBG_HIDRO_CON_EMBALSE}
   flg_flog_open:= false;
{$ENDIF}

end.
